from typing import Optional
import gdspy

from phidl.device_layout import DeviceReference

import pp
from pp.component import Component
from pp.name import NAME_TO_DEVICE
from pp.port import read_port_markers, auto_rename_ports
from pp.layers import port_layer2type, port_type2layer


def add_ports_from_markers_inside(*args, **kwargs):
    """markers inside the device"""
    return add_ports_from_markers_center(inside=True, *args, **kwargs)


def add_ports_from_markers_center(
    component: Component,
    port_layer2type=port_layer2type,
    port_type2layer=port_type2layer,
    inside: bool = False,
    tol=0.1,
):
    """add ports from polygons in certain layers

    markers at port center, so half of the marker goes inside and half ouside the port

    Args:
        component: to read polygons from and to write ports to
        port_layer2type: dict of layer to port_type
        port_type2layer: dict of port_type to layer
        inside: True-> markers  inside. False-> markers at center

    .. code::
           _______________
          |               |
          |               |
          |               |
         |||              |
         |||              |
          |               |
          |      __       |
          |_______________|
                 __

    dx < dy: port is east or west
        x > xc: east
        x < xc: west

    dx > dy: port is north or south
        y > yc: north
        y < yc: south

    dx = dy
        x > xc: east
        x < xc: west

    """
    i = 0
    xc = component.x
    yc = component.y
    xmax = component.xmax
    xmin = component.xmin
    ymax = component.ymax
    ymin = component.ymin

    for port_layer, port_type in port_layer2type.items():
        port_markers = read_port_markers(component, [port_layer])

        for p in port_markers.polygons:
            dy = p.ymax - p.ymin
            dx = p.xmax - p.xmin
            x = p.x
            y = p.y
            layer = port_type2layer[port_type]
            pxmax = p.xmax
            pxmin = p.xmin
            pymax = p.ymax
            pymin = p.ymin

            if dx < dy and x > xc:  # east
                orientation = 0
                width = dy
                if inside:
                    x = p.xmax
            elif dx < dy and x < xc:  # west
                orientation = 180
                width = dy
                if inside:
                    x = p.xmin
            elif dx > dy and y > yc:  # north
                orientation = 90
                width = dx
                if inside:
                    y = p.ymax
            elif dx > dy and y < yc:  # south
                orientation = 270
                width = dx
                if inside:
                    y = p.ymin
            # port markers have same width and height
            # check which edge (E, W, N, S) they are closer to
            elif pxmax > xmax - tol:  # east
                orientation = 0
                width = dy
                x = p.xmax
            elif pxmin < xmin + tol:  # west
                orientation = 180
                width = dy
                x = p.xmin
            elif pymax > ymax - tol:  # north
                orientation = 90
                width = dx
                y = p.ymax
            elif pymin < ymin + tol:  # south
                orientation = 270
                width = dx
                y = p.ymin
            else:
                raise ValueError(f"port marker x={x} y={y}, dx={dx}, dy={dy}")

            component.add_port(
                i,
                midpoint=(x, y),
                width=width,
                orientation=orientation,
                port_type=port_type,
                layer=layer,
            )
            i += 1


def import_gds_cells(gdspath):
    """ returns top cells from GDS"""
    gdsii_lib = gdspy.GdsLibrary()
    gdsii_lib.read_gds(gdspath)
    top_level_cells = gdsii_lib.top_level()
    return top_level_cells


def import_gds(
    gdspath: str,
    cellname: None = None,
    flatten: bool = False,
    overwrite_cache: bool = True,
    snap_to_grid_nm: Optional[int] = None,
) -> Component:
    """returns a Componenent from a GDS file

    Args:
        gdspath: path of GDS file
        cellname: cell of the name to import (None) imports top cell
        flatten: if True returns flattened (no hierarchy)
        overwrite_cache: overwrites device cache (caching by name)
        snap_to_grid_nm: snap

    """
    gdspath = str(gdspath)
    gdsii_lib = gdspy.GdsLibrary()
    gdsii_lib.read_gds(gdspath)
    top_level_cells = gdsii_lib.top_level()
    cellnames = [c.name for c in top_level_cells]

    if cellname is not None:
        if cellname not in cellnames:
            raise ValueError(
                f"import_gds() The requested cell {cellname} is not present in file {gdspath} with cells {cellnames}"
            )
        topcell = gdsii_lib.cells[cellname]
    elif cellname is None and len(top_level_cells) == 1:
        topcell = top_level_cells[0]
    elif cellname is None and len(top_level_cells) > 1:
        raise ValueError(
            "import_gds() There are multiple top-level cells in {}, you must specify `cellname` to select of one of them among {}".format(
                gdspath, [_c.name for _c in top_level_cells]
            )
        )

    if flatten:
        D = pp.Component()
        polygons = topcell.get_polygons(by_spec=True)

        for layer_in_gds, polys in polygons.items():
            D.add_polygon(polys, layer=layer_in_gds)
        return D

    else:
        D_list = []
        c2dmap = {}
        all_cells = topcell.get_dependencies(True)
        all_cells.update([topcell])

        for cell in all_cells:
            cell_name = cell.name
            if overwrite_cache or cell_name not in NAME_TO_DEVICE:
                D = pp.Component()
                D.name = cell.name
                D.polygons = cell.polygons
                D.references = cell.references
                D.name = cell_name
                D.labels = cell.labels
            else:
                D = NAME_TO_DEVICE[cell_name]

            c2dmap.update({cell_name: D})
            D_list += [D]

        for D in D_list:
            # First convert each reference so it points to the right Device
            converted_references = []
            for e in D.references:
                try:
                    ref_device = c2dmap[e.ref_cell.name]

                    dr = DeviceReference(
                        device=ref_device,
                        origin=e.origin,
                        rotation=e.rotation,
                        magnification=e.magnification,
                        x_reflection=e.x_reflection,
                    )
                    converted_references.append(dr)
                except Exception:
                    print("WARNING - Could not import", e.ref_cell.name)

            D.references = converted_references
            # Next convert each Polygon
            temp_polygons = list(D.polygons)
            D.polygons = []
            for p in temp_polygons:
                if snap_to_grid_nm:
                    points_on_grid = pp.drc.snap_to_grid(
                        p.polygons[0], nm=snap_to_grid_nm
                    )
                    p = gdspy.Polygon(
                        points_on_grid, layer=p.layers[0], datatype=p.datatypes[0]
                    )
                D.add_polygon(p)
                # else:
                #     warnings.warn('[PHIDL] import_gds(). Warning an element which was not a ' \
                #         'polygon or reference exists in the GDS, and was not able to be imported. ' \
                #         'The element was a: "%s"' % e)

        topdevice = c2dmap[topcell.name]
        return topdevice


def test_import_gds_snap_to_grid():
    gdspath = pp.CONFIG["gdsdir"] / "mmi1x2.gds"
    c = import_gds(gdspath, snap_to_grid_nm=5)
    print(len(c.get_polygons()))
    assert len(c.get_polygons()) == 8

    for x, y in c.get_polygons()[0]:
        assert pp.drc.on_grid(x, 5)
        assert pp.drc.on_grid(y, 5)


def test_import_gds_hierarchy():
    c0 = pp.c.mzi2x2()
    gdspath = pp.write_gds(c0)
    c = import_gds(gdspath)
    assert len(c.get_dependencies()) == 3


def test_import_gds_with_port_markers_optical():
    """ """
    # c  =  pp.c.mmi1x2()
    # for p in c.ports.values():
    #     print(p)
    # pp.show(c)

    gdspath = pp.CONFIG["gdsdir"] / "mmi1x2.gds"
    c = import_gds(gdspath)
    add_ports_from_markers_center(c)
    auto_rename_ports(c)

    # for p in c.ports.values():
    #     print(p)


def test_import_gds_with_port_markers_optical_electrical():
    """"""
    # c  =  pp.c.mzi2x2(with_elec_connections=True)
    # for p in c.ports.values():
    #     print(p)
    # pp.show(c)

    gdspath = pp.CONFIG["gdsdir"] / "mzi2x2.gds"
    c = import_gds(gdspath)
    add_ports_from_markers_center(c)
    auto_rename_ports(c)
    pp.show(c)

    for p in c.ports.values():
        print(p)


if __name__ == "__main__":
    test_import_gds_with_port_markers_optical_electrical()
    # test_import_gds_with_port_markers_optical()
    # test_import_gds_snap_to_grid()
    # test_import_gds_snap_to_grid()

    # gdspath = pp.CONFIG["gdslib"] / "gds" / "mzi2x2.gds"
    # c = import_gds(gdspath, snap_to_grid_nm=5)
    # print(c)
    # pp.show(c)



from pp.name import autoname
from pp import components as pc

import pp
from pp.component import Component
from typing import List


@autoname
def litho_steps(
    line_widths: List[int] = [1, 2, 4, 8, 16],
    line_spacing: int = 10,
    height: int = 100,
    layer: int = 0,
) -> Component:
    """ Produces a positive + negative tone linewidth test, used for
    lithography resolution test patterning

    Args:
        line_widths:
        line_spacing:
        height:
        layer:

    .. plot::
      :include-source:

      import pp

      c = pp.c.litho_steps()
      pp.plotgds(c)

    """
    D = pp.Component()

    height = height / 2
    T1 = pc.text(
        text="%s" % str(line_widths[-1]), size=height, justify="center", layer=layer
    )
    D.add_ref(T1).rotate(90).movex(-height / 10)
    R1 = pc.rectangle(size=(line_spacing, height), layer=layer)
    D.add_ref(R1).movey(-height)
    count = 0
    for i in reversed(line_widths):
        count += line_spacing + i
        R2 = pc.rectangle(size=(i, height), layer=layer)
        D.add_ref(R1).movex(count).movey(-height)
        D.add_ref(R2).movex(count - i)

    return D


if __name__ == "__main__":
    c = litho_steps()
    pp.show(c)



from typing import Any, Callable, List, Optional, Tuple, Union
import numpy as np
import phidl.device_layout as pd
from pp.routing.connect import get_waypoints_connect_strip
from pp.routing.utils import direction_ports_from_list_ports
from pp.routing.connect import connect_strip_way_points
from pp.routing.utils import flip
from pp.config import conf
from pp.component import Component, ComponentReference
from pp.port import Port


def route_south(
    component: Component,
    bend_radius: float = conf.tech.bend_radius,
    optical_routing_type: int = 1,
    excluded_ports: List[Any] = [],
    waveguide_separation: float = 4.0,
    io_gratings_lines: Optional[List[List[ComponentReference]]] = None,
    route_filter: Callable = connect_strip_way_points,
    gc_port_name: str = "E0",
) -> Union[Tuple[List[Any], List[Port]], Tuple[List[ComponentReference], List[Port]]]:
    """
    Args:
        component: component to route
        bend_radius
        optical_routing_type: routing heuristic ``1`` or ``2`` (see below)
        excluded_ports=[]: list of port names to NOT route
        waveguide_separation
        io_gratings_lines: list of ports to which the ports produced by this
            function will be connected. Supplying this information helps
            avoiding waveguide collisions

        routing_method: routing method to connect the waveguides
        gc_port_name: grating port name

    Returns:
        list of elements, list of ports

    Standard optical routing - type ``1`` or variant ``2``
        ``1`` uses the component size info to estimate the box size
        ``2`` only looks at the optical port positions to estimate the size

    Works well if the component looks rougly like a rectangular box with
        north ports on the north of the box
        south ports on the south of the box
        east ports on the east of the box
        west ports on the west of the box
    """

    optical_ports = component.get_ports_list(port_type="optical")
    optical_ports = [p for p in optical_ports if p.name not in excluded_ports]
    csi = component.size_info
    elements = []

    # Handle empty list gracefully
    if not optical_ports:
        return [], []

    conn_params = {"bend_radius": bend_radius}

    route_filter_params = {
        "bend_radius": bend_radius,
        "wg_width": optical_ports[0].width,
    }

    def routing_method(p1, p2, **kwargs):
        way_points = get_waypoints_connect_strip(p1, p2, **kwargs)
        return route_filter(way_points, **route_filter_params)

    # Used to avoid crossing between waveguides in special cases
    # This could happen when abs(x_port - x_grating) <= 2 * bend_radius
    delta_gr_min = 2 * bend_radius + 1

    sep = waveguide_separation

    # Get lists of optical ports by orientation
    direction_ports = direction_ports_from_list_ports(optical_ports)

    north_ports = direction_ports["N"]
    north_start = north_ports[0 : len(north_ports) // 2]
    north_finish = north_ports[len(north_ports) // 2 :]

    west_ports = direction_ports["W"]
    west_ports.reverse()
    east_ports = direction_ports["E"]
    south_ports = direction_ports["S"]
    north_finish.reverse()  # Sort right to left
    north_start.reverse()  # Sort right to left
    ordered_ports = north_start + west_ports + south_ports + east_ports + north_finish

    def get_index_port_closest_to_x(x, list_ports):
        return np.array([abs(x - p.ports[gc_port_name].x) for p in list_ports]).argmin()

    def gen_port_from_port(x, y, p):
        new_p = pd.Port(name=p.name, midpoint=(x, y), orientation=90.0, width=p.width)

        return new_p

    R = bend_radius
    west_ports.reverse()

    y0 = min([p.y for p in ordered_ports]) - R - 0.5

    ports_to_route = []

    i = 0
    optical_xs_tmp = [p.x for p in ordered_ports]
    x_optical_min = min(optical_xs_tmp)
    x_optical_max = max(optical_xs_tmp)

    """
    ``x`` is the x-coord of the waypoint where the current component port is connected.
    x starts as close as possible to the component.
    For each new port, the distance is increased by the separation.
    The starting x depends on the heuristic chosen : ``1`` or ``2``
    """

    # Set starting ``x`` on the west side
    if optical_routing_type == 1:
        ## `` use component size to know how far to route
        x = csi.west - R - 1
    elif optical_routing_type == 2:
        ## `` use optical port to know how far to route
        x = x_optical_min - R - 1
    else:
        raise ValueError("Invalid optical routing type")

    # First route the ports facing west
    for p in west_ports:
        """
        In case we have to connect these ports to a line of grating,
        Ensure that the port is aligned with the grating port or
        has enough space for manhattan routing (at least two bend radius)
        """
        if io_gratings_lines:
            i_grating = get_index_port_closest_to_x(x, io_gratings_lines[-1])
            x_gr = io_gratings_lines[-1][i_grating].ports[gc_port_name].x
            if abs(x - x_gr) < delta_gr_min:
                if x > x_gr:
                    x = x_gr
                elif x < x_gr:
                    x = x_gr - delta_gr_min

        tmp_port = gen_port_from_port(x, y0, p)
        ports_to_route.append(tmp_port)
        elements += [routing_method(p, tmp_port, **conn_params)]
        x -= sep

        i += 1
    start_straight = 0.5

    # First-half of north ports
    # This ensures that north ports are routed above the top west one
    north_start.reverse()  # We need them from left to right
    if len(north_start) > 0:
        y_max = max([p.y for p in west_ports + north_start])
        for p in north_start:
            tmp_port = gen_port_from_port(x, y0, p)

            elements += [
                routing_method(
                    p,
                    tmp_port,
                    start_straight=start_straight + y_max - p.y,
                    **conn_params
                )
            ]

            ports_to_route.append(tmp_port)
            x -= sep
            start_straight += sep

    # Set starting ``x`` on the east side
    if optical_routing_type == 1:
        ## `` use component size to know how far to route
        x = csi.east + R + 1
    elif optical_routing_type == 2:
        ## `` use optical port to know how far to route
        x = x_optical_max + R + 1
    else:
        raise ValueError("Invalid optical routing type")
    i = 0

    # Route the east ports
    start_straight = 0.5
    for p in east_ports:
        """
        In case we have to connect these ports to a line of grating,
        Ensure that the port is aligned with the grating port or
        has enough space for manhattan routing (at least two bend radius)
        """
        if io_gratings_lines:
            i_grating = get_index_port_closest_to_x(x, io_gratings_lines[-1])
            x_gr = io_gratings_lines[-1][i_grating].ports[gc_port_name].x
            if abs(x - x_gr) < delta_gr_min:
                if x < x_gr:
                    x = x_gr
                elif x > x_gr:
                    x = x_gr + delta_gr_min

        tmp_port = gen_port_from_port(x, y0, p)

        elements += [
            routing_method(p, tmp_port, start_straight=start_straight, **conn_params)
        ]

        ports_to_route.append(tmp_port)
        x += sep
        i += 1

    # Route the remaining north ports
    start_straight = 0.5
    if len(north_finish) > 0:
        y_max = max([p.y for p in east_ports + north_finish])
        for p in north_finish:
            tmp_port = gen_port_from_port(x, y0, p)
            ports_to_route.append(tmp_port)
            elements += [
                routing_method(
                    p,
                    tmp_port,
                    start_straight=start_straight + y_max - p.y,
                    **conn_params
                )
            ]
            x += sep
            start_straight += sep

    # Add south ports
    ports = [flip(p) for p in ports_to_route] + south_ports

    return elements, ports


if __name__ == "__main__":
    import pp

    c = pp.c.mmi2x2()
    elements, ports = route_south(c)
    for e in elements:
        c.add(e)

    pp.show(c)



from collections import defaultdict
import math
import klayout.db as pya
import pp.autoplacer.functions as ap


class CellList:
    """ just a list of components,
    with the ability to normalize size and shape """

    def __init__(self, cells):
        self.cells = cells

    def find_origin(self, cell):
        """ finds the displacement of the origin """
        origins = []
        for instance in cell.each_inst():
            if instance.cell.name.endswith("t_c"):
                origins.append(instance.trans.disp)

        if origins:
            return pya.Point(
                ap.mean([o.x for o in origins]), ap.mean([o.y for o in origins])
            )
        else:
            bbox = cell.bbox()
            return pya.Point(-bbox.left, -bbox.bottom)

    def align(self, dimension=ap.BOTH):
        """ Aligns the grating couplers """
        if not self.cells:
            return

        for cell in self.cells:
            cell.origin = self.find_origin(cell)

        max_origin = pya.Point(
            max(cell.origin.x for cell in self.cells),
            max(cell.origin.y for cell in self.cells),
        )

        for cell in self.cells:
            delta = max_origin - cell.origin
            layer = cell.layout().layer(ap.DEVREC_LAYER, 0)
            delta.x = delta.x if dimension in (ap.BOTH, ap.WIDTH) else 0
            delta.y = delta.y if dimension in (ap.BOTH, ap.HEIGHT) else 0
            box = cell.bbox().enlarge(delta.x, delta.y)
            cell.shapes(layer).insert(box)

    def normalize(self, dimension, origin=ap.SOUTH_WEST):
        """ Makes all components the same size on some dimension """
        if not self.cells:
            return

        max_width = max(cell.bbox().width() for cell in self.cells)
        max_height = max(cell.bbox().height() for cell in self.cells)
        for cell in self.cells:
            bbox = cell.bbox()
            if not (bbox.width() == max_width and bbox.height == max_height):
                dx = max_width - bbox.width() if dimension in (ap.WIDTH, ap.BOTH) else 0
                dy = (
                    max_height - bbox.height()
                    if dimension in (ap.HEIGHT, ap.BOTH)
                    else 0
                )

                if origin == ap.SOUTH_WEST:
                    box = pya.Box(
                        bbox.left, bbox.bottom, bbox.right + dx, bbox.top + dy
                    )
                elif origin == ap.SOUTH_EAST:
                    box = pya.Box(
                        bbox.left - dx, bbox.bottom, bbox.right, bbox.top + dy
                    )
                elif origin == ap.NORTH_WEST:
                    box = pya.Box(
                        bbox.left, bbox.bottom - dy, bbox.right + dx, bbox.top
                    )
                elif origin == ap.NORTH_EAST:
                    box = pya.Box(
                        bbox.left - dx, bbox.bottom - dy, bbox.right, bbox.top
                    )

                layer = cell.layout().layer(ap.DEVREC_LAYER, 0)
                cell.shapes(layer).insert(box)

    def pad(self, padding=ap.PADDING):
        """ add padding to a cell """
        for cell in self.cells:
            layer = cell.layout().layer(ap.DEVREC_LAYER, 0)
            bbox = cell.bbox()
            cell.shapes(layer).insert(bbox)
            box = pya.Box(
                bbox.left - padding,
                bbox.bottom - padding,
                bbox.right + padding,
                bbox.top + padding,
            )
            cell.shapes(layer).insert(box)

    def sort(self, key):
        """ Sort the cells """
        self.cells = sorted(self.cells, key=key)

    def groups(self, dimension, granularity=ap.GRANULARITY):
        """ group cells by size """
        groups = defaultdict(list)
        for cell in self.cells:
            b = cell.bbox()
            lookup = {ap.WIDTH: b.width(), ap.HEIGHT: b.height(), ap.AREA: b.area()}
            size = lookup[dimension]
            size = granularity * math.floor(size // granularity)
            groups[size].append(cell)
        return [
            CellList(sorted(groups[key], key=ap.area))
            for key in sorted(list(groups.keys()), reverse=True)
        ]

    def __iter__(self):
        """ allow for cell in celllist """
        for cell in sorted(self.cells, key=ap.area):
            yield cell

    def __getitem__(self, index):
        """ allow for cell in celllist """
        return self.cells[index]

    def __len__(self):
        return len(self.cells)

    def __str__(self):
        s = "\n".join(" - " + c.name for c in self.cells[:3])
        if len(self.cells) > 3:
            s += "\n..."
        return "list of {} cells:\n{}".format(len(self.cells), s)



from pp.components import taper
from pp.components import compass
from pp.name import autoname
from pp.component import Component
import pp
from typing import Tuple


@autoname
def grating_coupler_uniform(
    num_teeth: int = 20,
    period: float = 0.75,
    fill_factor: float = 0.5,
    width_grating: float = 11.0,
    length_taper: float = 150.0,
    width: float = 0.5,
    partial_etch: bool = False,
    layer: Tuple[int, int] = pp.LAYER.WG,
    layer_partial_etch: Tuple[int, int] = pp.LAYER.SLAB150,
) -> Component:
    """ Grating coupler uniform

    Args:
        num_teeth: 20
        period: 0.75
        fill_factor: 0.5
        width_grating: 11
        length_taper: 150
        width: 0.5
        partial_etch: False

    .. plot::
      :include-source:

      import pp

      c = pp.c.grating_coupler_uniform()
      pp.plotgds(c)

    """
    # returns a fiber grating
    G = Component()

    if partial_etch:
        partetch_overhang = 5
        _compass = compass(
            size=[period * (1 - fill_factor), width_grating + partetch_overhang * 2],
            layer=layer_partial_etch,
        )

        # make the etched areas (opposite to teeth)
        for i in range(num_teeth):
            cgrating = G.add_ref(_compass)
            cgrating.x += i * period

        # draw the deep etched square around the grating
        deepbox = G.add_ref(
            compass(size=[num_teeth * period, width_grating], layer=layer)
        )
        deepbox.movex(num_teeth * period / 2)
    else:
        for i in range(num_teeth):
            cgrating = G.add_ref(
                compass(size=[period * fill_factor, width_grating], layer=layer)
            )
            cgrating.x += i * period
    # make the taper
    tgrating = G.add_ref(
        taper(
            length=length_taper,
            width1=width_grating,
            width2=width,
            port=None,
            layer=layer,
        )
    )
    tgrating.xmin = cgrating.xmax
    # define the port of the grating
    G.add_port(port=tgrating.ports["2"], name="E0")
    return G


def test_grating_coupler_uniform():
    c = grating_coupler_uniform(name="gcu", partial_etch=False)
    pp.write_gds(c)


if __name__ == "__main__":
    # c = grating_coupler_uniform(name='gcu', partial_etch=True)
    c = grating_coupler_uniform(partial_etch=False)
    pp.show(c)



import pp
from pp.components.pcm.cd import CENTER_SHAPES_MAP


@pp.autoname
def cd_straight(
    spacing_h=5.0,
    spacing_v=8.0,
    gaps=[0.224, 0.234, 0.246],
    length=10.0,
    width_center=0.5,
    layer=pp.LAYER.WG,
):

    c = pp.Component()
    x = 0
    i = 0

    widths = [width_center * 0.92, width_center, width_center * 1.08]

    for width, marker_type in zip(widths, ["D", "S", "U"]):
        y = 0
        # iso line
        _r = pp.c.rectangle_centered(x=length, y=width, layer=layer)
        _r_ref = c.add_ref(_r)
        _r_ref.move((x, y))
        c.absorb(_r_ref)

        marker = CENTER_SHAPES_MAP[marker_type]()
        _marker = c.add_ref(marker)
        _marker.move((x, -2.0))
        c.absorb(_marker)

        y += width + spacing_v
        # dual lines
        for gap, col_marker_type in zip(gaps, ["S", "D", "U"]):
            _r1_ref = c.add_ref(_r)
            _r1_ref.move((x, y))

            _r2_ref = c.add_ref(_r)
            _r2_ref.move((x, y + gap + width))
            c.absorb(_r1_ref)
            c.absorb(_r2_ref)

            if i < 2:
                marker = CENTER_SHAPES_MAP[col_marker_type]()
                _marker = c.add_ref(marker)
                _marker.move((x + length / 2 + spacing_h / 2, y + width + gap / 2))
                c.absorb(_marker)

            y += 2 * width + gap + spacing_v
        i += 1

        x += length + spacing_h

    c.move(c.size_info.cc, (0, 0))
    return c


if __name__ == "__main__":
    c = cd_straight()
    pp.show(c)



from typing import Callable
from pp.components.mmi1x2 import mmi1x2
import pp


@pp.autoname
def splitter_chain(component: Callable = mmi1x2, n_devices: int = 3):
    """ Chain of splitters

    .. code::

                __
             __|
          __|  |__
        _|  |__
         |__


    .. plot::
      :include-source:

      import pp

      c = pp.c.splitter_chain(component=pp.c.mmi1x2(), n_devices=3)
      pp.plotgds(c)

    """
    c = pp.Component()
    component = pp.component_factory(component)
    cref = c.add_ref(component)

    bend = pp.c.bezier()
    c.add_port(name="W0", port=cref.ports["W0"])
    c.add_port(name="E0", port=cref.ports["E0"])

    for i in range(1, n_devices):
        bref = c.add_ref(bend)
        bref.connect(port="1", destination=cref.ports["E1"])

        component = pp.call_if_func(component)
        cref = c.add_ref(component)

        cref.connect(port="W0", destination=bref.ports["0"])
        c.add_port(name="E{}".format(i), port=cref.ports["E0"])

    c.add_port(name="E{}".format(n_devices), port=cref.ports["E1"])
    return c


if __name__ == "__main__":
    c = splitter_chain(component=pp.c.mmi1x2, n_devices=4)
    pp.show(c)



from pp import klive
from pp.autoplacer.library import Library
from pp.autoplacer.chip_array import ChipArray


def test_autplacer():
    lib = Library()
    mask = ChipArray("chip_array", 25e6, 25e6, 3, 4, lib)
    mask.pack_grid(lib.pop("align"))
    mask.pack_grid(lib.pop(".*"))
    mask.write("chip_array.gds")
    klive.show("chip_array.gds")


if __name__ == "__main__":
    test_autplacer()



import pp
from pp.components.grating_coupler.elliptical_trenches import grating_coupler_te
from pp.components.grating_coupler.elliptical_trenches import grating_coupler_tm
from pp.routing.get_input_labels import get_input_labels
from pp.container import container


@container
def add_grating_couplers(
    component,
    grating_coupler=grating_coupler_te,
    layer_label=pp.LAYER.LABEL,
    gc_port_name="W0",
    get_input_labels_function=get_input_labels,
):
    """ returns component with grating ports and labels on each port
    """
    component = pp.call_if_func(component)
    c = pp.Component(name=component.name + "_c")
    c.add_ref(component)
    grating_coupler = pp.call_if_func(grating_coupler)

    io_gratings = []
    for i, port in enumerate(component.ports.values()):
        gc_ref = grating_coupler.ref()
        gc_ref.connect(list(gc_ref.ports.values())[0], port)
        io_gratings.append(gc_ref)
        c.add(gc_ref)

    labels = get_input_labels_function(
        io_gratings,
        list(component.ports.values()),
        component_name=component.name,
        layer_label=layer_label,
        gc_port_name=gc_port_name,
    )
    c.add(labels)

    return c


def add_te(*args, **kwargs):
    return add_grating_couplers(*args, **kwargs)


def add_tm(*args, grating_coupler=grating_coupler_tm, **kwargs):
    return add_grating_couplers(*args, grating_coupler=grating_coupler, **kwargs)


if __name__ == "__main__":
    # from pp.add_labels import get_optical_text
    # c = pp.c.grating_coupler_elliptical_te()
    # print(c.wavelength)

    # print(c.get_property('wavelength'))

    c = pp.c.waveguide(width=2)
    # cc = add_grating_couplers(c)
    cc = add_tm(c)
    print(cc)
    pp.show(cc)



import numpy as np

from phidl.geometry import _width
from phidl.geometry import _indent
from phidl.geometry import _glyph

import pp
from pp.layers import LAYER
from pp.components.manhattan_font import manhattan_text
from pp.name import clean_name
from pp.component import Component
from typing import Tuple


def text(
    text: str = "abcd",
    size: float = 10.0,
    position: Tuple[int, int] = (0, 0),
    justify: str = "left",
    layer: Tuple[int, int] = LAYER.TEXT,
) -> Component:
    """ adds text

    .. plot::
      :include-source:

      import pp

      c = pp.c.text(text="abcd", size=10, position=(0, 0), justify="left", layer=1)
      pp.plotgds(c)

    """
    scaling = size / 1000
    xoffset = position[0]
    yoffset = position[1]
    t = pp.Component(
        name=clean_name(text) + "_{}_{}".format(int(position[0]), int(position[1]))
    )
    for i, line in enumerate(text.split("\n")):
        label = pp.Component(name=t.name + "{}".format(i))
        for c in line:
            ascii_val = ord(c)
            if c == " ":
                xoffset += 500 * scaling
            elif 33 <= ascii_val <= 126:
                for poly in _glyph[ascii_val]:
                    xpts = np.array(poly)[:, 0] * scaling
                    ypts = np.array(poly)[:, 1] * scaling
                    label.add_polygon([xpts + xoffset, ypts + yoffset], layer=layer)
                xoffset += (_width[ascii_val] + _indent[ascii_val]) * scaling
            else:
                ValueError(
                    "[PHIDL] text(): No glyph for character with ascii value %s"
                    % ascii_val
                )
        t.add_ref(label)
        yoffset -= 1500 * scaling
        xoffset = position[0]
    justify = justify.lower()
    for label in t.references:
        if justify == "left":
            pass
        if justify == "right":
            label.xmax = position[0]
        if justify == "center":
            label.move(origin=label.center, destination=position, axis="x")
    return t


@pp.autoname
def githash(text=[], size=0.4, hash_length=6, layer=LAYER.WG):
    """ returns the photonics_pdk git hash
    allows a list of text, that will print on separate lines ::

    text = [
        "sw_{}".format(Repo(CONFIG["repo"]).head.object.hexsha[:length]),
        "ap_{}".format(Repo(ap.CONFIG["repo"]).head.object.hexsha[:length]),
        "mm_{}".format(Repo(mm.CONFIG["repo"]).head.object.hexsha[:length]),
    ]
    c = githash(text=text)
    pp.write_gds(c)
    pp.show(c)

    """
    try:
        git_hash = "pp_{}".format(pp.CONFIG["repo"][:hash_length])
    except Exception:
        git_hash = "pp_{}".format(pp.__version__)

    c = pp.Component()
    t = manhattan_text(text=git_hash, size=size, layer=layer)
    tref = c.add_ref(t)
    c.absorb(tref)

    for i, texti in enumerate(text):
        t = manhattan_text(text=texti, size=size, layer=layer)
        tref = c.add_ref(t)
        tref.movey(-6 * size * (i + 1))
        c.absorb(tref)
    return c


if __name__ == "__main__":
    c = text(
        text=".[,ABCDEFGHIKKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789:/",
        size=4.0,
        position=(120.5, 3),
    )
    c = githash(text=["a", "b"], size=10)
    pp.show(c)



"""
"""

import pp
from pp.routing import route_manhattan
from pp.components.mmi1x2 import mmi1x2
from pp.components.euler.bend_euler import bend_euler90
from pp.components.spiral_external_io import spiral_external_io
from pp.component import Component
from typing import Callable


@pp.autoname
def loop_mirror(
    component: Callable = mmi1x2, bend90: Callable = bend_euler90
) -> Component:
    c = pp.Component()
    component = pp.call_if_func(component)
    bend90 = pp.call_if_func(bend90)
    cref = c.add_ref(component)
    elements = route_manhattan(
        cref.ports["E0"],
        cref.ports["E1"],
        bend90=bend90,
        straight_factory=pp.c.waveguide,
    )
    c.add(elements)
    c.add_port(name="W0", port=cref.ports["W0"])
    c.absorb(cref)
    return c


@pp.autoname
def loop_mirror_rotated(component=mmi1x2, bend90=bend_euler90):
    c = pp.Component()
    component = pp.call_if_func(component)
    mirror = loop_mirror(component=component, bend90=bend90)
    mirror_rotated = mirror.ref(rotation=90)
    c.add(mirror_rotated)
    c.absorb(mirror_rotated)
    c.add_port(name="S0", port=mirror_rotated.ports["W0"])
    return c


@pp.autoname
def loop_mirror_with_delay(loop_mirror=loop_mirror, spiral=spiral_external_io):
    """
    delay = 13e-12
    # delay = length/speed
    # length=delay*speed
    13e-12*3e8/4.2*1e6

    """
    c = pp.Component()
    lm = c << pp.call_if_func(loop_mirror)
    s = c << pp.call_if_func(spiral_external_io)

    lm.connect("W0", s.ports["input"])
    return c


if __name__ == "__main__":
    # c = loop_mirror()
    # c = loop_mirror_rotated()
    c = loop_mirror_with_delay()
    pp.show(c)



import pp
from pp.components.grating_coupler.elliptical_trenches import grating_coupler_te
from pp.components.grating_coupler.elliptical_trenches import grating_coupler_tm
from pp.routing.utils import check_ports_have_equal_spacing
from pp.routing.utils import direction_ports_from_list_ports
from pp.routing.manhattan import round_corners
from pp.components import waveguide
from pp.components import bend_circular
from pp.components.taper import taper
from pp.add_labels import get_input_label
from pp.container import container


@container
def add_termination(component, terminator=taper):
    """ returns component containing a comonent with all ports terminated """
    terminator = pp.call_if_func(terminator)
    c = pp.Component(name=component.name + "_t")
    c.add_ref(component)

    for port in component.ports.values():
        t_ref = c.add_ref(terminator)
        t_ref.connect(list(t_ref.ports.values())[0].name, port)

    return c


def add_gratings_and_loop_back_te(*args, **kwargs):
    return add_gratings_and_loop_back(*args, **kwargs)


def add_gratings_and_loop_back_tm(*args, grating_coupler=grating_coupler_tm, **kwargs):
    return add_gratings_and_loop_back(*args, grating_coupler=grating_coupler, **kwargs)


@container
def add_gratings_and_loop_back(
    component,
    grating_coupler=grating_coupler_te,
    excluded_ports=[],
    grating_separation=127.0,
    bend_radius_align_ports=10.0,
    gc_port_name=None,
    gc_rotation=-90,
    waveguide_separation=5.0,
    bend_factory=bend_circular,
    waveguide_factory=waveguide,
    layer_label=pp.LAYER.LABEL,
    component_name=None,
    with_loopback=True,
):
    """ returns a component with grating_couplers and loopback
    """
    gc = pp.call_if_func(grating_coupler)

    direction = "S"
    component_name = component_name or component.name
    name = f"{component.name}_{gc.polarization}"
    c = pp.Component(name=name)
    c.add_ref(component)

    # Find grating port name if not specified
    if gc_port_name is None:
        gc_port_name = list(gc.ports.values())[0].name

    # List the optical ports to connect
    optical_ports = component.get_ports_list(port_type="optical")
    optical_ports = [p for p in optical_ports if p.name not in excluded_ports]
    optical_ports = direction_ports_from_list_ports(optical_ports)[direction]

    # Check if the ports are equally spaced
    grating_separation_extracted = check_ports_have_equal_spacing(optical_ports)
    if grating_separation_extracted != grating_separation:
        raise ValueError(
            "Grating separation must be {}. Got {}".format(
                grating_separation, grating_separation_extracted
            )
        )

    # Add grating couplers
    couplers = []
    for port in optical_ports:
        coupler_ref = c.add_ref(gc)
        coupler_ref.connect(list(coupler_ref.ports.values())[0].name, port)
        couplers += [coupler_ref]

    # add labels
    for i, optical_port in enumerate(optical_ports):
        label = get_input_label(
            optical_port,
            couplers[i],
            i,
            component_name=component_name,
            layer_label=layer_label,
        )
        c.add(label)

    if with_loopback:
        y0 = couplers[0].ports[gc_port_name].y
        xs = [p.x for p in optical_ports]
        x0 = min(xs) - grating_separation
        x1 = max(xs) + grating_separation

        gca1, gca2 = [
            gc.ref(position=(x, y0), rotation=gc_rotation, port_id=gc_port_name)
            for x in [x0, x1]
        ]

        gsi = gc.size_info
        p0 = gca1.ports[gc_port_name].position
        p1 = gca2.ports[gc_port_name].position
        a = bend_radius_align_ports + 0.5
        b = max(2 * a, grating_separation / 2)
        y_bot_align_route = -gsi.width - waveguide_separation

        route = [
            p0,
            p0 + (0, a),
            p0 + (b, a),
            p0 + (b, y_bot_align_route),
            p1 + (-b, y_bot_align_route),
            p1 + (-b, a),
            p1 + (0, a),
            p1,
        ]
        bend90 = bend_factory(radius=bend_radius_align_ports)
        loop_back = round_corners(route, bend90, waveguide_factory)
        elements = [gca1, gca2, loop_back]
        c.add(elements)
    return c


if __name__ == "__main__":
    # gc = pp.c.grating_coupler_elliptical_te()
    # cc = add_termination(c, gc)
    from pp.components.spiral_inner_io import spiral_inner_io

    c = spiral_inner_io()
    # c = waveguide()
    cc = add_gratings_and_loop_back(c, with_loopback=False)

    # cc = add_termination(component=c)
    print(cc.get_settings()["component"])
    pp.show(cc)



import datetime
import platform
import pp
from pp.layers import LAYER
from pp.components.text import text as Text


def Pixel(size=1.0, layer=LAYER.WG):
    c = pp.Component()
    a = size / 2
    c.add_polygon([(a, a), (a, -a), (-a, -a), (-a, a)], layer)
    return c


@pp.autoname
def qrcode(data, psize=1, layer=LAYER.WG):
    """ A utility to render a list of pixels as a block """
    import qrcode

    pixel = Pixel(size=psize, layer=layer)
    q = qrcode.QRCode()
    q.add_data(data)
    matrix = q.get_matrix()
    c = pp.Component()
    for i, row in enumerate(matrix):
        for j, value in enumerate(row):
            if value:
                _p = pixel.ref((i * psize, j * psize))
                c.add(_p)
                c.absorb(_p)
    return c


@pp.autoname
def version_stamp(
    text=[], with_qr_code=False, layer=LAYER.WG, pixel_size=1, text_size=10
):
    """
    Includes module version, git hash and date

    .. plot::
      :include-source:

      import pp

      c = pp.c.version_stamp()
      pp.plotgds(c)

    """

    git_hash = pp.CONFIG["git_hash"]
    now = datetime.datetime.now()
    timestamp = "{:%Y-%m-%d %H:%M:%S}".format(now)
    short_stamp = "{:%y.%m.%d.%H.%M.%S}".format(now)

    c = pp.Component()
    if with_qr_code:
        data = "{}/{}/{}".format(git_hash, timestamp, platform.node())
        q = qrcode(layer=layer, data=data, psize=pixel_size).ref_center()
        c.add(q)
        c.absorb(q)

        x = q.size_info.width * 0.5 + 10

    else:
        x = 0

    txt_params = {"layer": layer, "justify": "left", "size": text_size}
    date = Text(
        position=(x, text_size + 2 * pixel_size), text=short_stamp, **txt_params
    ).ref()
    c.add(date)
    c.absorb(date)

    git_hash = Text(position=(x, 0), text=git_hash[:15], **txt_params).ref()
    c.add(git_hash)
    c.absorb(git_hash)

    for i, line in enumerate(text):
        text = c << Text(
            position=(x, -(i + 1) * (text_size + 2 * pixel_size)),
            text=line,
            **txt_params
        )
        c.absorb(text)

    return c


if __name__ == "__main__":
    print(pp.CONFIG["git_hash"])
    c = version_stamp(
        pixel_size=4,
        layer=LAYER.M1,
        with_qr_code=True,
        # text=["b1", "demo"],
        text_size=20,
    )
    pp.show(c)



from typing import Callable
from pp.components.bend_circular import bend_circular
from pp.components.coupler_ring import coupler_ring
from pp.components.waveguide import waveguide
from pp.drc import assert_on_2nm_grid
from pp.component import Component
from pp.config import call_if_func
from pp.name import autoname


@autoname
def ring_single(
    wg_width: float = 0.5,
    gap: float = 0.2,
    length_x: float = 4.0,
    bend_radius: float = 5.0,
    length_y: float = 2.0,
    coupler: Callable = coupler_ring,
    waveguide: Callable = waveguide,
    bend: Callable = bend_circular,
) -> Component:
    """ single bus ring made of two couplers (ct: top, cb: bottom)
    connected with two vertical waveguides (wyl: left, wyr: right)

    .. code::

          bl-wt-br
          |      |
          wl     wr length_y
          |      |
         --==cb==-- gap

          length_x

    .. plot::
      :include-source:

      import pp

      c = pp.c.ring_single(wg_width=0.5, gap=0.2, length_x=4, length_y=0.1, bend_radius=5)
      pp.plotgds(c)

    """
    bend_radius = float(bend_radius)
    assert_on_2nm_grid(gap)

    coupler = call_if_func(
        coupler, gap=gap, wg_width=wg_width, bend_radius=bend_radius, length_x=length_x
    )
    waveguide_side = call_if_func(waveguide, width=wg_width, length=length_y)
    waveguide_top = call_if_func(waveguide, width=wg_width, length=length_x)
    bend = call_if_func(bend, width=wg_width, radius=bend_radius)

    c = Component()
    cb = c << coupler
    wl = c << waveguide_side
    wr = c << waveguide_side
    bl = c << bend
    br = c << bend
    wt = c << waveguide_top

    wl.connect(port="E0", destination=cb.ports["N0"])
    bl.connect(port="N0", destination=wl.ports["W0"])

    wt.connect(port="W0", destination=bl.ports["W0"])
    br.connect(port="N0", destination=wt.ports["E0"])
    wr.connect(port="W0", destination=br.ports["W0"])
    wr.connect(port="E0", destination=cb.ports["N1"])  # just for netlist

    c.add_port("E0", port=cb.ports["E0"])
    c.add_port("W0", port=cb.ports["W0"])
    return c


if __name__ == "__main__":
    import pp

    c = ring_single(pins=True)
    # print(c.settings)
    # print(c.get_settings())
    pp.show(c)



import sys
import collections
from multiprocessing import Process
import time
from pprint import pprint
from omegaconf import OmegaConf

from pp.placer import save_doe
from pp.placer import doe_exists
from pp.placer import _gen_components
from pp.placer import _gen_components_from_generator
from pp.placer import load_doe_component_names

from pp.config import CONFIG
from pp.components import component_type2factory
from pp.write_doe import write_doe_metadata
from pp.doe import get_settings_list

from pp.config import logging


def _print(*args, **kwargs):
    print(*args, **kwargs)
    sys.stdout.flush()


def _pprint(*args, **kwargs):
    pprint(*args, **kwargs)
    sys.stdout.flush()


def _test(doe_params):
    print(doe_params[0])


def default_component_filter(x):
    return x


def separate_does_from_templates(dicts):
    type_to_dict = {}

    does = {}
    for name, d in dicts.items():
        if "type" in d.keys():
            template_type = d.pop("type")
            if template_type not in type_to_dict:
                type_to_dict[template_type] = {}
            type_to_dict[template_type][name] = d
        else:
            does[name] = d

    return does, type_to_dict


def update_dicts_recurse(target_dict, default_dict):
    target_dict = target_dict.copy()
    default_dict = default_dict.copy()
    for k, v in default_dict.items():
        if k not in target_dict:
            target_dict[k] = v
        else:
            if isinstance(target_dict[k], (dict, collections.OrderedDict)):
                target_dict[k] = update_dicts_recurse(target_dict[k], default_dict[k])
    return target_dict


def save_doe_use_template(doe, doe_root_path=None):
    """
    Write a "content.txt" pointing to the DOE used as a template
    """
    doe_name = doe["name"]
    doe_template = doe["doe_template"]
    doe_root_path = doe_root_path or CONFIG["cache_doe_directory"]
    doe_dir = doe_root_path / doe_name
    doe_dir.mkdir(exist_ok=True)
    content_file = doe_dir / "content.txt"

    with open(content_file, "w") as fw:
        fw.write("TEMPLATE: {}".format(doe_template))


def _generate_doe(
    doe,
    component_type2factory=component_type2factory,
    component_filter=default_component_filter,
    doe_root_path=None,
    doe_metadata_path=None,
    regenerate_report_if_doe_exists=False,
    precision=1e-9,
    logger=logging,
    **kwargs,
):
    doe_name = doe["name"]
    list_settings = doe["list_settings"]

    line = "Building - {} ...".format(doe_name)
    logger.info(line)

    # Otherwise generate each component using the component factory
    component_type = doe["component"]
    if "generator" in doe:
        components = _gen_components_from_generator(
            doe["generator"],
            component_type,
            doe,
            component_type2factory=component_type2factory,
        )
    else:
        components = _gen_components(
            component_type, list_settings, component_type2factory=component_type2factory
        )

    components = [component_filter(c) for c in components]
    component_names = [c.name for c in components]
    save_doe(doe_name, components, doe_root_path=doe_root_path, precision=precision)

    write_doe_metadata(
        doe_name=doe["name"],
        cell_names=component_names,
        list_settings=doe["list_settings"],
        doe_settings=kwargs,
        doe_metadata_path=doe_metadata_path,
    )


def load_does(filepath, defaults={"do_permutation": True, "settings": {}}):
    does = {}
    data = OmegaConf.load(filepath)
    data = OmegaConf.to_container(data)
    mask = data.pop("mask")

    for doe_name, doe in data.items():
        for k in defaults:
            if k not in doe:
                doe[k] = defaults[k]

        does[doe_name] = doe
    return does, mask


def generate_does(
    filepath,
    component_filter=default_component_filter,
    component_type2factory=component_type2factory,
    doe_root_path=CONFIG["cache_doe_directory"],
    doe_metadata_path=CONFIG["doe_directory"],
    n_cores=4,
    logger=logging,
    regenerate_report_if_doe_exists=False,
    precision=1e-9,
):
    """ Generates a DOEs of components specified in a yaml file
    allows for each DOE to have its own x and y spacing (more flexible than method1)
    similar to write_doe
    """

    doe_root_path.mkdir(parents=True, exist_ok=True)
    doe_metadata_path.mkdir(parents=True, exist_ok=True)

    dicts, mask_settings = load_does(filepath)
    does, templates_by_type = separate_does_from_templates(dicts)

    dict_templates = (
        templates_by_type["template"] if "template" in templates_by_type else {}
    )

    default_use_cached_does = (
        mask_settings["cache"] if "cache" in mask_settings else False
    )

    list_args = []
    for doe_name, doe in does.items():
        doe["name"] = doe_name
        component = doe["component"]

        if component not in component_type2factory:
            raise ValueError(f"{component} not in {component_type2factory.keys()}")

        if "template" in doe:
            """
            The keyword template is used to enrich the dictionnary from the template
            """
            templates = doe["template"]
            if not isinstance(templates, list):
                templates = [templates]
            for template in templates:
                try:
                    doe = update_dicts_recurse(doe, dict_templates[template])
                except Exception:
                    print(template, "does not exist")
                    raise

        do_permutation = doe.pop("do_permutation")
        settings = doe["settings"]
        doe["list_settings"] = get_settings_list(do_permutation, **settings)

        list_args += [doe]

    does_running = []
    start_times = {}
    finish_times = {}
    doe_name_to_process = {}
    while list_args:
        while len(does_running) < n_cores:
            if not list_args:
                break
            doe = list_args.pop()
            doe_name = doe["name"]

            """
            Only launch a build process if we do not use the cache
            Or if the DOE is not built
            """

            list_settings = doe["list_settings"]

            use_cached_does = (
                default_use_cached_does if "cache" not in doe else doe["cache"]
            )

            _doe_exists = False

            if "doe_template" in doe:
                """
                In that case, the DOE is not built: this DOE points to another existing component
                """
                _doe_exists = True
                logger.info("Using template - {}".format(doe_name))
                save_doe_use_template(doe)

            elif use_cached_does:
                _doe_exists = doe_exists(doe_name, list_settings)
                if _doe_exists:
                    logger.info("Cached - {}".format(doe_name))
                    if regenerate_report_if_doe_exists:
                        component_names = load_doe_component_names(doe_name)

                        write_doe_metadata(
                            doe_name=doe["name"],
                            cell_names=component_names,
                            list_settings=doe["list_settings"],
                            doe_metadata_path=doe_metadata_path,
                        )

            if not _doe_exists:
                start_times[doe_name] = time.time()
                p = Process(
                    target=_generate_doe,
                    args=(doe, component_type2factory),
                    kwargs={
                        "component_filter": component_filter,
                        "doe_root_path": doe_root_path,
                        "doe_metadata_path": doe_metadata_path,
                        "regenerate_report_if_doe_exists": regenerate_report_if_doe_exists,
                        "precision": precision,
                        "logger": logger,
                    },
                )
                doe_name_to_process[doe_name] = p
                does_running += [doe_name]
                try:
                    p.start()
                except Exception:
                    print("Issue starting process for {}".format(doe_name))
                    print(type(component_type2factory))
                    raise

        to_rm = []
        for i, doe_name in enumerate(does_running):
            _p = doe_name_to_process[doe_name]
            if not _p.is_alive():
                to_rm += [i]
                finish_times[doe_name] = time.time()
                dt = finish_times[doe_name] - start_times[doe_name]
                line = "Done - {} ({:.1f}s)".format(doe_name, dt)
                logger.info(line)

        for i in to_rm[::-1]:
            does_running.pop(i)

        time.sleep(0.001)

    while does_running:
        to_rm = []
        for i, _doe_name in enumerate(does_running):
            _p = doe_name_to_process[_doe_name]
            if not _p.is_alive():
                to_rm += [i]
        for i in to_rm[::-1]:
            does_running.pop(i)

        time.sleep(0.05)


if __name__ == "__main__":
    filepath = CONFIG["samples_path"] / "mask" / "does.yml"
    generate_does(filepath, precision=2e-9)



import itertools
import uuid
import copy as python_copy
import pathlib
from typing import Any, Dict, List, Optional, Tuple, Union
import numpy as np
from numpy import float64, int64, ndarray, pi, sin, cos, mod
from omegaconf import OmegaConf
import networkx as nx

from phidl.device_layout import Label
from phidl.device_layout import Device
from phidl.device_layout import DeviceReference
from phidl.device_layout import _parse_layer

from pp.port import Port, select_ports
from pp.config import CONFIG, conf, connections
from pp.compare_cells import hash_cells
from pp.name import dict2hash, clean_dict, clean_list


def copy(D):
    """returns a copy of a Component.
    """
    D_copy = Component(name=D._internal_name)
    D_copy.info = python_copy.deepcopy(D.info)
    for ref in D.references:
        new_ref = ComponentReference(
            ref.parent,
            origin=ref.origin,
            rotation=ref.rotation,
            magnification=ref.magnification,
            x_reflection=ref.x_reflection,
        )
        new_ref.owner = D_copy
        D_copy.add(new_ref)
        for alias_name, alias_ref in D.aliases.items():
            if alias_ref == ref:
                D_copy.aliases[alias_name] = new_ref

    for port in D.ports.values():
        D_copy.add_port(port=port)
    for poly in D.polygons:
        D_copy.add_polygon(poly)
    for label in D.labels:
        D_copy.add_label(
            text=label.text,
            position=label.position,
            layer=(label.layer, label.texttype),
        )
    return D_copy


class SizeInfo:
    def __init__(self, bbox: ndarray) -> None:
        self.west = bbox[0, 0]
        self.east = bbox[1, 0]
        self.south = bbox[0, 1]
        self.north = bbox[1, 1]

        self.width = self.east - self.west
        self.height = self.north - self.south

        xc = 0.5 * (self.east + self.west)
        yc = 0.5 * (self.north + self.south)

        self.sw = np.array([self.west, self.south])
        self.se = np.array([self.east, self.south])
        self.nw = np.array([self.west, self.north])
        self.ne = np.array([self.east, self.north])

        self.cw = np.array([self.west, yc])
        self.ce = np.array([self.east, yc])
        self.nc = np.array([xc, self.north])
        self.sc = np.array([xc, self.south])
        self.cc = self.center = np.array([xc, yc])

    def get_rect(
        self, padding=0, padding_w=None, padding_e=None, padding_n=None, padding_s=None
    ):
        w, e, s, n = self.west, self.east, self.south, self.north

        padding_n = padding if padding_n is None else padding_n
        padding_e = padding if padding_e is None else padding_e
        padding_w = padding if padding_w is None else padding_w
        padding_s = padding if padding_s is None else padding_s

        w = w - padding_w
        e = e + padding_e
        s = s - padding_s
        n = n + padding_n

        return [(w, s), (e, s), (e, n), (w, n)]

    @property
    def rect(self):
        return self.get_rect()

    def __str__(self):
        return "w: {}\ne: {}\ns: {}\nn: {}\n".format(
            self.west, self.east, self.south, self.north
        )


def _rotate_points(
    points: Union[Tuple[int, int], ndarray],
    angle: Union[float64, int, int64, float] = 45,
    center: Union[Tuple[int, int], List[int], ndarray] = (0, 0),
) -> ndarray:
    """Rotates points around a centerpoint defined by ``center``.  ``points`` may be
    input as either single points [1,2] or array-like[N][2], and will return in kind
    """
    # First check for common, easy values of angle
    p_arr = np.asarray(points)
    if angle == 0:
        return p_arr

    c0 = np.asarray(center)
    displacement = p_arr - c0
    if angle == 180:
        return c0 - displacement

    if p_arr.ndim == 2:
        perpendicular = displacement[:, ::-1]
    elif p_arr.ndim == 1:
        perpendicular = displacement[::-1]

    # Fall back to trigonometry
    angle = angle * pi / 180
    ca = cos(angle)
    sa = sin(angle)
    sa = np.array((-sa, sa))
    return displacement * ca + perpendicular * sa + c0


class ComponentReference(DeviceReference):
    def __init__(
        self,
        component,
        origin: Tuple[int, int] = (0, 0),
        rotation: int = 0,
        magnification: None = None,
        x_reflection: bool = False,
        visual_label: str = "",
    ) -> None:
        super().__init__(
            device=component,
            origin=origin,
            rotation=rotation,
            magnification=magnification,
            x_reflection=x_reflection,
        )
        self.parent = component
        # The ports of a DeviceReference have their own unique id (uid),
        # since two DeviceReferences of the same parent Device can be
        # in different locations and thus do not represent the same port
        self._local_ports = {
            name: port._copy(new_uid=True) for name, port in component.ports.items()
        }
        self.visual_label = visual_label

    def __repr__(self):
        return (
            'DeviceReference (parent Device "%s", ports %s, origin %s, rotation %s,'
            " x_reflection %s)"
            % (
                self.parent.name,
                list(self.ports.keys()),
                self.origin,
                self.rotation,
                self.x_reflection,
            )
        )

    def __str__(self):
        return self.__repr__()

    def __getitem__(self, val):
        """This allows you to access an alias from the reference's parent, and receive
        a copy of the reference which is correctly rotated and translated"""
        try:
            alias_device = self.parent[val]
        except Exception:
            raise ValueError(
                '[PHIDL] Tried to access alias "%s" from parent '
                'Device "%s", which does not exist' % (val, self.parent.name)
            )
        new_reference = ComponentReference(
            alias_device.parent,
            origin=alias_device.origin,
            rotation=alias_device.rotation,
            magnification=alias_device.magnification,
            x_reflection=alias_device.x_reflection,
        )

        if self.x_reflection:
            new_reference.reflect((1, 0))
        if self.rotation is not None:
            new_reference.rotate(self.rotation)
        if self.origin is not None:
            new_reference.move(self.origin)

        return new_reference

    def get_labels(
        self, recursive: None = True, associate_visual_labels: bool = True
    ) -> List[Any]:
        """
        access all labels correctly rotated, mirrored and translated
        """
        # params = {
        #     "recursive": recursive,
        #     "associate_visual_labels": associate_visual_labels,
        # }

        labels_untransformed = self.parent.get_labels()

        labels = []
        visual_label = self.visual_label
        for lbl in labels_untransformed:
            p = lbl.position
            position, _ = self._transform_port(
                p, 0, self.origin, self.rotation, self.x_reflection
            )

            text = lbl.text
            if visual_label and associate_visual_labels:
                text += " / " + visual_label

            new_lbl = Label(
                text=text,
                position=position,
                anchor="o",
                layer=lbl.layer,
                texttype=lbl.texttype,
            )

            labels += [new_lbl]
        if labels:
            print("R", self.parent.name, len(labels))
        return labels

    @property
    def ports(self) -> Dict[str, Port]:
        """This property allows you to access myref.ports, and receive a copy
        of the ports dict which is correctly rotated and translated"""
        for name, port in self.parent.ports.items():
            port = self.parent.ports[name]
            new_midpoint, new_orientation = self._transform_port(
                port.midpoint,
                port.orientation,
                self.origin,
                self.rotation,
                self.x_reflection,
            )
            if name not in self._local_ports:
                self._local_ports[name] = port._copy(new_uid=True)
            self._local_ports[name].midpoint = new_midpoint
            self._local_ports[name].orientation = mod(new_orientation, 360)
            self._local_ports[name].parent = self
        # Remove any ports that no longer exist in the reference's parent
        parent_names = self.parent.ports.keys()
        local_names = list(self._local_ports.keys())
        for name in local_names:
            if name not in parent_names:
                self._local_ports.pop(name)
        return self._local_ports

    @property
    def info(self) -> Dict[str, Union[float64, float]]:
        return self.parent.info

    @property
    def size_info(self) -> SizeInfo:
        return SizeInfo(self.bbox)
        # if self.__size_info__ == None:
        # return self.__size_info__

    def _transform_port(
        self,
        point: ndarray,
        orientation: Union[float64, int64],
        origin: Union[Tuple[int, int], ndarray] = (0, 0),
        rotation: Optional[Union[float64, int, int64]] = None,
        x_reflection: bool = False,
    ) -> Union[Tuple[ndarray, float64], Tuple[ndarray, int64]]:
        # Apply GDS-type transformations to a port (x_ref)
        new_point = np.array(point)
        new_orientation = orientation

        if x_reflection:
            new_point[1] = -new_point[1]
            new_orientation = -orientation
        if rotation is not None:
            new_point = _rotate_points(new_point, angle=rotation, center=[0, 0])
            new_orientation += rotation
        if origin is not None:
            new_point = new_point + np.array(origin)
        new_orientation = mod(new_orientation, 360)

        return new_point, new_orientation

    def _transform_point(
        self,
        point: ndarray,
        origin: Union[Tuple[int, int], ndarray] = (0, 0),
        rotation: Optional[Union[int64, int, float]] = None,
        x_reflection: bool = False,
    ) -> ndarray:
        # Apply GDS-type transformations to a port (x_ref)
        new_point = np.array(point)

        if x_reflection:
            new_point[1] = -new_point[1]
        if rotation is not None:
            new_point = _rotate_points(new_point, angle=rotation, center=[0, 0])
        if origin is not None:
            new_point = new_point + np.array(origin)

        return new_point

    def move(
        self,
        origin: Union[
            Port,
            ndarray,
            List[Union[float, float64]],
            Tuple[int, int],
            List[Union[int, float]],
        ] = (0, 0),
        destination: Optional[Any] = None,
        axis: Optional[str] = None,
    ):
        """Moves the DeviceReference from the origin point to the destination.  Both
        origin and destination can be 1x2 array-like, Port, or a key
        corresponding to one of the Ports in this device_ref

        Returns:
            ComponentReference
        """

        # If only one set of coordinates is defined, make sure it's used to move things
        if destination is None:
            destination = origin
            origin = (0, 0)

        if isinstance(origin, Port):
            o = origin.midpoint
        elif np.array(origin).size == 2:
            o = origin
        elif origin in self.ports:
            o = self.ports[origin].midpoint
        else:
            raise ValueError(
                "[ComponentReference.move()] ``origin`` not array-like, a port, or port"
                " name"
            )

        if isinstance(destination, Port):
            d = destination.midpoint
        elif np.array(destination).size == 2:
            d = destination
        elif destination in self.ports:
            d = self.ports[destination].midpoint
        else:
            raise ValueError(
                "[ComponentReference.move()] ``destination`` not array-like, a port, or"
                " port name"
            )

        # Lock one axis if necessary
        if axis == "x":
            d = (d[0], o[1])
        if axis == "y":
            d = (o[0], d[1])

        # This needs to be done in two steps otherwise floating point errors can accrue
        dxdy = np.array(d) - np.array(o)
        self.origin = np.array(self.origin) + dxdy
        self._bb_valid = False
        return self

    def rotate(
        self, angle: [int, float] = 45, center: Tuple[int, int] = (0.0, 0.0),
    ):
        """
        Returns a component
            ComponentReference
        """
        if angle == 0:
            return self
        if type(center) == str or type(center) == int:
            center = self.ports[center].position

        if type(center) is Port:
            center = center.midpoint
        self.rotation += angle
        self.rotation = self.rotation % 360
        self.origin = _rotate_points(self.origin, angle, center)
        self._bb_valid = False
        return self

    def reflect_h(self, port_name=None, x0=None):
        """
        Perform horizontal mirror (w.r.t vertical axis)
        """
        if port_name is None and x0 is None:
            x0 = 0

        if port_name is not None:
            position = self.ports[port_name]
            x0 = position.x
        self.reflect((x0, 1), (x0, 0))

    def reflect_v(self, port_name: Optional[str] = None, y0: None = None) -> None:
        """
        Perform vertical mirror (w.r.t horizontal axis)
        """
        if port_name is None and y0 is None:
            y0 = 0

        if port_name is not None:
            position = self.ports[port_name]
            y0 = position.y
        self.reflect((1, y0), (0, y0))

    def reflect(
        self,
        p1: Union[Tuple[float64, float64], Tuple[int, float64]] = (0, 1),
        p2: Union[Tuple[float64, float64], Tuple[int, float64]] = (0, 0),
    ):
        if isinstance(p1, Port):
            p1 = p1.midpoint
        if isinstance(p2, Port):
            p2 = p2.midpoint
        p1 = np.array(p1)
        p2 = np.array(p2)
        # Translate so reflection axis passes through origin
        self.origin = self.origin - p1

        # Rotate so reflection axis aligns with x-axis
        angle = np.arctan2((p2[1] - p1[1]), (p2[0] - p1[0])) * 180 / pi
        self.origin = _rotate_points(self.origin, angle=-angle, center=[0, 0])
        self.rotation -= angle

        # Reflect across x-axis
        self.x_reflection = not self.x_reflection
        self.origin[1] = -self.origin[1]
        self.rotation = -self.rotation

        # Un-rotate and un-translate
        self.origin = _rotate_points(self.origin, angle=angle, center=[0, 0])
        self.rotation += angle
        self.rotation = self.rotation % 360
        self.origin = self.origin + p1

        self._bb_valid = False
        return self

    def connect(self, port: str, destination: Port, overlap: float = 0):
        """returns ComponentReference"""
        # ``port`` can either be a string with the name or an actual Port
        if port in self.ports:  # Then ``port`` is a key for the ports dict
            p = self.ports[port]
        elif type(port) is Port:
            p = port
        else:
            raise ValueError(
                f"{self}.connect({port}) valid ports are {self.ports.keys()}"
            )

        angle = 180 + destination.orientation - p.orientation
        angle = angle % 360
        self.rotate(angle=angle, center=p.midpoint)

        self.move(origin=p, destination=destination)
        self.move(
            -overlap
            * np.array(
                [
                    cos(destination.orientation * pi / 180),
                    sin(destination.orientation * pi / 180),
                ]
            )
        )
        if destination.parent:
            global connections
            # connections[f"{self.parent.uid}_{int(self.x)}_{int(self.y)},{port}"] = f"{destination.parent.get_property('uid')},{destination.name}"
            connections[
                f"{self.get_property('name')}_{int(self.x)}_{int(self.y)},{port}"
            ] = f"{destination.parent.get_property('name')}_{int(destination.parent.x)}_{int(destination.parent.y)},{destination.name}"
        return self

    def get_property(self, property: str) -> Union[str, int]:
        if hasattr(self, property):
            return self.property

        return self.ref_cell.get_property(property)


class Component(Device):
    """adds some functions to phidl.Device

    - get/write JSON metadata
    - get ports by type (optical, electrical ...)
    - set data_analysis and test_protocols

    """

    def __init__(
        self,
        name: str = "Unnamed",
        polarization: None = None,
        wavelength: None = None,
        test_protocol: None = None,
        data_analysis_protocol: None = None,
        *args,
        **kwargs,
    ) -> None:
        # Allow name to be set like Component('arc') or Component(name = 'arc')

        self.data_analysis_protocol = data_analysis_protocol or {}
        self.test_protocol = test_protocol or {}
        self.wavelength = wavelength
        self.polarization = polarization

        self.settings = kwargs
        self.__ports__ = {}
        self.info = {}
        self.aliases = {}
        self.uid = str(uuid.uuid4())[:8]

        if "with_uuid" in kwargs or name == "Unnamed":
            name += "_" + self.uid

        super(Component, self).__init__(name=name, exclude_from_current=True)
        self.name = name
        self.name_long = None
        self.function_name = None

    def plot_netlist(
        self, label_index_end=1, with_labels=True, font_weight="normal",
    ):
        """plots a netlist graph with networkx
        https://networkx.github.io/documentation/stable/reference/generated/networkx.drawing.nx_pylab.draw_networkx.html

        Args:
            label_index_end: name args separated with `_` to print (-1: all)
            with_labels: label nodes
            font_weight: normal, bold
        """
        import matplotlib.pyplot as plt

        netlist = self.get_netlist()
        connections = netlist.connections
        G = nx.Graph()
        G.add_edges_from(
            [(k.split(",")[0], v.split(",")[0]) for k, v in connections.items()]
        )
        pos = {k: (v["x"], v["y"]) for k, v in netlist.placements.items()}
        labels = {
            k: "_".join(k.split("_")[:label_index_end])
            for k in netlist.placements.keys()
        }
        nx.draw(
            G, with_labels=with_labels, font_weight=font_weight, labels=labels, pos=pos
        )
        plt.show()

    def get_netlist(self, full_settings=False):
        """returns netlist dict(instances, placements, connections)
        if full_settings: exports all the settings
        """
        instances = {}
        placements = {}

        for r in self.references:
            i = r.parent
            reference_name = f"{i.name}_{int(r.x)}_{int(r.y)}"
            if full_settings:
                settings = i.settings
            else:
                settings = i.settings_changed
            instances[reference_name] = dict(
                component=i.function_name, settings=settings
            )
            placements[reference_name] = dict(
                x=float(r.x), y=float(r.y), rotation=int(r.rotation)
            )

        connections_connected = {}

        # print(instances.keys())
        for src, dst in connections.items():
            # print(src.split(',')[0])
            # trim netlist:
            # only instances that are part of the component are connected
            if src.split(",")[0] in instances:
                connections_connected[src] = dst

        netlist = OmegaConf.create(
            dict(
                instances=instances,
                placements=placements,
                connections=connections_connected,
            )
        )
        self.netlist = netlist
        return netlist

    def get_name_long(self):
        """ returns the long name if it's been truncated to MAX_NAME_LENGTH"""
        if self.name_long:
            return self.name_long
        else:
            return self.name

    def get_sparameters_path(self, dirpath=CONFIG["sp"], height_nm=220):
        dirpath = pathlib.Path(dirpath)
        dirpath = dirpath / self.function_name if self.function_name else dirpath
        dirpath.mkdir(exist_ok=True, parents=True)
        return dirpath / f"{self.get_name_long()}_{height_nm}.dat"

    def ports_on_grid(self) -> None:
        """ asserts if all ports ar eon grid """
        for port in self.ports.values():
            port.on_grid()

    def get_ports_dict(self, port_type="optical", prefix=None):
        """ returns a list of ports """
        return select_ports(self.ports, port_type=port_type, prefix=prefix)

    def get_ports_list(self, port_type="optical", prefix=None) -> List[Port]:
        """ returns a lit of  ports """
        return list(
            select_ports(self.ports, port_type=port_type, prefix=prefix).values()
        )

    def get_ports_array(self) -> Dict[str, ndarray]:
        """ returns ports as a dict of np arrays"""
        self.ports_on_grid()
        ports_array = {
            port_name: np.array(
                [
                    port.x,
                    port.y,
                    int(port.orientation),
                    port.width,
                    port.layer[0],
                    port.layer[1],
                ]
            )
            for port_name, port in self.ports.items()
        }
        return ports_array

    def get_properties(self):
        """ returns name, uid, ports, aliases and numer of references """
        return (
            f"name: {self._internal_name}, uid: {self.uid},  ports:"
            f" {self.ports.keys()}, aliases {self.aliases.keys()}, number of"
            f" references: {len(self.references)}"
        )

    def ref(
        self,
        position: Union[
            Tuple[float, float], Port, Tuple[int, float], ndarray, Tuple[int, int]
        ] = (0, 0),
        port_id: Optional[str] = None,
        rotation: Union[float64, int, int64] = 0,
        h_mirror: bool = False,
        v_mirror: bool = False,
    ) -> ComponentReference:
        """ returns a reference of the component """
        _ref = ComponentReference(self)

        if port_id and port_id not in self.ports:
            raise ValueError(f"port {port_id} not in {self.ports.keys()}")

        if port_id:
            origin = self.ports[port_id].position
        else:
            origin = (0, 0)

        if h_mirror:
            _ref.reflect_h(port_id)

        if v_mirror:
            _ref.reflect_v(port_id)

        if rotation != 0:
            _ref.rotate(rotation, origin)
        _ref.move(origin, position)
        return _ref

    def ref_center(self, position=(0, 0)):
        """ returns a reference of the component centered at (x=0, y=0)"""
        si = self.size_info
        yc = si.south + si.height / 2
        xc = si.west + si.width / 2
        center = (xc, yc)
        _ref = ComponentReference(self)
        _ref.move(center, position)
        return _ref

    def __repr__(self) -> str:
        return self.name

    def __str__(self):
        return self.__repr__()

    def update_settings(self, **kwargs):
        """ update settings dict """
        for key, value in kwargs.items():
            self.settings[key] = _clean_value(value)

    def get_property(self, property: str) -> Union[str, int]:
        if property in self.settings:
            return self.settings[property]
        if hasattr(self, property):
            return getattr(self, property)

        raise ValueError(
            "Component {} does not have property {}".format(self.name, property)
        )

    def get_settings(self) -> Dict[str, Any]:
        """Returns settings dictionary"""
        output = {}
        ignore = set(
            dir(Component())
            + [
                "path",
                "settings",
                "properties",
                "function_name",
                "type",
                "netlist",
                "pins",
                "settings_changed",
            ]
        )
        params = set(dir(self)) - ignore
        output["name"] = self.name

        if hasattr(self, "function_name"):
            output["function_name"] = self.function_name

        for key, value in self.settings.items():
            output[key] = _clean_value(value)

        for param in params:
            value = getattr(self, param)
            output[param] = _clean_value(value)

        # output["hash"] = hashlib.md5(json.dumps(output).encode()).hexdigest()
        # output["hash_geometry"] = str(self.hash_geometry())
        output = {k: output[k] for k in sorted(output)}
        return output

    def get_settings_model(self):
        """ returns important settings for a compact model"""
        ignore = ["layer", "layers_cladding", "cladding_offset"]
        s = self.get_settings()
        [s.pop(i) for i in ignore]
        return s

    def add_port(
        self,
        name: Optional[Union[str, int]] = None,
        midpoint: Any = (0, 0),
        width: Union[float64, int, float] = 1,
        orientation: Union[int, int64, float] = 45,
        port: Optional[Port] = None,
        layer: Union[Tuple[int, int], int] = (1, 0),
        port_type: str = "optical",
    ) -> Port:
        """Can be called to copy an existing port like add_port(port = existing_port) or
        to create a new port add_port(myname, mymidpoint, mywidth, myorientation).
        Can also be called to copy an existing port with a new name like add_port(port = existing_port, name = new_name)"""
        if port:
            if not isinstance(port, Port):
                print(type(port))
                raise ValueError(
                    "[PHIDL] add_port() error: Argument `port` must be a Port for"
                    " copying"
                )
            p = port._copy(new_uid=True)
            if name is not None:
                p.name = name
            p.parent = self

        elif isinstance(name, Port):
            p = name._copy(new_uid=True)
            p.parent = self
            name = p.name
        else:
            p = Port(
                name=name,
                midpoint=midpoint,
                width=width,
                orientation=orientation,
                parent=self,
                layer=layer,
                port_type=port_type,
            )
        if name is not None:
            p.name = name
        if p.name in self.ports:
            raise ValueError(
                '[DEVICE] add_port() error: Port name "%s" already exists in this'
                ' Device (name "%s", uid %s)' % (p.name, self._internal_name, self.uid)
            )

        self.ports[p.name] = p
        return p

    def snap_ports_to_grid(self, nm=1):
        for port in self.ports.values():
            port.snap_to_grid(nm=nm)

    def get_json(self, **kwargs) -> Dict[str, Any]:
        """ returns JSON metadata """
        jsondata = {
            "json_version": 7,
            "cells": recurse_structures(self),
            "test_protocol": self.test_protocol,
            "data_analysis_protocol": self.data_analysis_protocol,
            "git_hash": conf["git_hash"],
        }

        if hasattr(self, "analysis"):
            jsondata["analysis"] = self.analysis

        return jsondata

    def __hash__(self) -> int:
        h = dict2hash(**self.settings)
        return int(h, 16)

    def hash_geometry(self):
        """returns geometrical hash"""
        if self.references or self.polygons:
            h = hash_cells(self, {})[self.name]
        else:
            h = "empty_geometry"

        self.settings.update(hash=h)
        return h

    def remove_layers(
        self, layers=(), include_labels=True, invert_selection=False, recursive=True
    ):
        """ remove a list of layers """
        layers = [_parse_layer(layer) for layer in layers]
        all_D = list(self.get_dependencies(recursive))
        all_D += [self]
        for D in all_D:
            for polygonset in D.polygons:
                polygon_layers = zip(polygonset.layers, polygonset.datatypes)
                polygons_to_keep = [(pl in layers) for pl in polygon_layers]
                if not invert_selection:
                    polygons_to_keep = [(not p) for p in polygons_to_keep]
                polygonset.polygons = [
                    p for p, keep in zip(polygonset.polygons, polygons_to_keep) if keep
                ]
                polygonset.layers = [
                    p for p, keep in zip(polygonset.layers, polygons_to_keep) if keep
                ]
                polygonset.datatypes = [
                    p for p, keep in zip(polygonset.datatypes, polygons_to_keep) if keep
                ]

            if include_labels:
                new_labels = []
                for label in D.labels:
                    original_layer = (label.layer, label.texttype)
                    original_layer = _parse_layer(original_layer)
                    if invert_selection:
                        keep_layer = original_layer in layers
                    else:
                        keep_layer = original_layer not in layers
                    if keep_layer:
                        new_labels += [label]
                D.labels = new_labels
        return self

    def copy(self):
        return copy(self)

    @property
    def size_info(self) -> SizeInfo:
        """ size info of the component """
        # if self.__size_info__ == None:
        # self.__size_info__  = SizeInfo(self.bbox)
        return SizeInfo(self.bbox)  # self.__size_info__

    def add_ref(self, D, alias: Optional[str] = None) -> ComponentReference:
        """Takes a Component and adds it as a ComponentReference to the current
        Device."""
        if type(D) in (list, tuple):
            return [self.add_ref(E) for E in D]
        if not isinstance(D, Component) and not isinstance(D, Device):
            raise TypeError(
                """[PP] add_ref() was passed a {}. This is not a Component object. """.format(
                    type(D)
                )
            )
        d = ComponentReference(D)  # Create a ComponentReference (CellReference)
        self.add(d)  # Add ComponentReference (CellReference) to Device (Cell)

        if alias is not None:
            self.aliases[alias] = d
        return d

    def get_layers(self):
        """returns a set of (layer, datatype)

        >>> import pp
        >>> pp.c.waveguide().get_layers() == {(1, 0), (111, 0)}

        """
        layers = set()
        for element in itertools.chain(self.polygons, self.paths):
            for layer, datatype in zip(element.layers, element.datatypes):
                layers.add((layer, datatype))
        for reference in self.references:
            for layer, datatype in reference.ref_cell.get_layers():
                layers.add((layer, datatype))
        for label in self.labels:
            layers.add((label.layer, 0))
        return layers


def test_get_layers():
    import pp

    c = pp.c.waveguide()
    assert c.get_layers() == {(1, 0), (111, 0)}
    c.remove_layers((111, 0))
    assert c.get_layers() == {(1, 0)}


def _filter_polys(polygons, layers_excl):
    return [
        p
        for p, l, d in zip(polygons.polygons, polygons.layers, polygons.datatypes)
        if (l, d) not in layers_excl
    ]


IGNORE_FUNCTION_NAMES = set()
IGNORE_STRUCTURE_NAME_PREFIXES = set(["zz_conn"])


def recurse_structures(structure: Component) -> Dict[str, Any]:
    """ Recurse over structures """
    if (
        hasattr(structure, "function_name")
        and structure.function_name in IGNORE_FUNCTION_NAMES
    ):
        return {}

    if hasattr(structure, "name") and any(
        [structure.name.startswith(i) for i in IGNORE_STRUCTURE_NAME_PREFIXES]
    ):
        return {}
    if not hasattr(structure, "get_json"):
        return {}

    # if hasattr(structure, "function_name"):
    #     print(structure.function_name)
    output = {structure.name: structure.get_settings()}
    for element in structure.references:
        if (
            isinstance(element, ComponentReference)
            and element.ref_cell.name not in output
        ):
            output.update(recurse_structures(element.ref_cell))

    return output


def _clean_value(value: Any) -> Any:
    """ returns a clean value """
    if type(value) in [int, float, str, tuple, bool]:
        value = value
    elif isinstance(value, np.int32):
        value = int(value)
    elif isinstance(value, np.int64):
        value = int(value)
    elif callable(value):
        value = value.__name__
    elif type(value) == Component:
        value = value.name
    elif hasattr(value, "items"):
        clean_dict(value)
    elif hasattr(value, "__iter__"):
        clean_list(value)
    else:
        value = str(value)

    return value


def test_same_uid():
    import pp

    c = Component()
    c << pp.c.rectangle()
    c << pp.c.rectangle()

    r1 = c.references[0].parent
    r2 = c.references[1].parent

    print(r1.uid, r2.uid)
    print(r1 == r2)


def test_netlist_simple():
    import pp

    c = pp.Component()
    c1 = c << pp.c.waveguide(length=1, width=1)
    c2 = c << pp.c.waveguide(length=2, width=2)
    c2.connect(port="W0", destination=c1.ports["E0"])
    c.add_port("W0", port=c1.ports["W0"])
    c.add_port("E0", port=c2.ports["E0"])
    netlist = c.get_netlist()
    # print(netlist.pretty())
    assert len(netlist["instances"]) == 2
    assert len(netlist["connections"]) == 1


def test_netlist_complex():
    import pp

    c = pp.c.mzi()
    netlist = c.get_netlist()
    # print(netlist.pretty())
    assert len(netlist["instances"]) == 18
    assert len(netlist["connections"]) == 18


def demo_component(port):
    c = Component()
    c.add_port(name="p1", port=port)
    return c


if __name__ == "__main__":
    import pp

    # c = pp.c.ring_single()
    # c = pp.c.mzi()
    # c.plot_netlist()

    # test_netlist_simple()
    # test_netlist_complex()

    # c = pp.c.waveguide()
    # c = pp.c.dbr(n=1)

    # print(c.get_layers())

    # c = pp.c.bend_circular180()
    # c = pp.c.coupler()
    # c.add_labels()
    # pp.show(c)
    # test_same_uid()

    # c = pp.c.mmi1x2()
    # c = pp.c.mzi1x2()

    # print(c.hash_geometry())
    # print(c.get_json())
    # print(c.get_settings())
    # print(c.settings)

    # print(c.get_settings())
    # print(c.get_ports_array())

    # print(json.dumps(c.get_settings()))
    # print(c.get_json()['cells'].keys())
    # print(c.get_json())

    # from pp.routing import add_fiber_array
    # cc = add_fiber_array(c)
    # pp.write_component(cc)

    # from pprint import pprint

    # c = pp.c.mmi1x2()
    # cc = add_fiber_array(c)
    # cc.get_json()
    # pp.show(cc)
    # c.update_settings(
    #     analysis={
    #         "device_type": "loopback",
    #         "test_method": "spectrum",
    #         "analysis_keys": "[cell_name, category, 0]",
    #         "polarization_mix": "None",
    #         "device_group_type": "deembed, mapping",
    #         "optical_io_measure": "[[1,4],[2,3]]",
    #         "polarization_input": "TE",
    #         "deembed_device_group_id": "None",
    #     }
    # )

    # w = pp.c.waveguide()
    # c = demo_component(port=w.ports["E0"])
    # pprint(c.get_json())
    # pprint(c.get_settings())

    c = pp.c.waveguide()
    c = pp.routing.add_fiber_array(c)
    print(c.get_settings())
    # print(c.get_json())
    # print(c.get_settings(test="hi"))



from numpy import pi, cos, sin
import numpy as np

import pp
from pp.layers import LAYER
from pp.component import Component
from typing import List, Tuple, Union


def _interpolate_segment(p0, p1, N=2):
    p0 = np.array(p0)
    p1 = np.array(p1)
    dp = p1 - p0
    d = np.sqrt(dp[0] ** 2 + dp[1] ** 2)
    return [p0 + dp * a / d for a in np.linspace(0, d, N)]


def _bend_path_from_pts(pts, n_interp=2):
    N = len(pts)
    pts0 = pts[: N // 2]
    pts1 = pts[N // 2 :][::-1]

    pts = [((x0 + x1) * 0.5, (y0 + y1) * 0.5) for (x0, y0), (x1, y1) in zip(pts0, pts1)]

    all_pts = [pts[0]]
    for p0, p1 in zip(pts[:-1], pts[1:]):
        _pts = _interpolate_segment(p0, p1, N=n_interp)
        all_pts += _pts[1:]

    return all_pts


def _bend_path(radius=10.0, start_angle=0, theta=-90, angle_resolution=2.5):
    angle1 = (start_angle) * pi / 180
    angle2 = (start_angle + theta) * pi / 180
    t = np.linspace(angle1, angle2, int(abs(theta) / angle_resolution))
    points_x = (radius * cos(t)).tolist()
    points_y = (radius * sin(t)).tolist()
    return points_x, points_y


def _bend_points(
    radius=10.0,
    width=0.5,
    theta=-90,
    start_angle=0,
    angle_resolution=2.5,
    inner_radius=None,
    outer_radius=None,
):
    inner_radius = inner_radius or radius - width / 2
    outer_radius = outer_radius or radius + width / 2
    angle1 = (start_angle) * pi / 180
    angle2 = (start_angle + theta) * pi / 180
    t = np.linspace(angle1, angle2, int(abs(theta) / angle_resolution))
    inner_points_x = (inner_radius * cos(t)).tolist()
    inner_points_y = (inner_radius * sin(t)).tolist()
    outer_points_x = (outer_radius * cos(t)).tolist()
    outer_points_y = (outer_radius * sin(t)).tolist()
    xpts = inner_points_x + outer_points_x[::-1]
    ypts = inner_points_y + outer_points_y[::-1]
    return xpts, ypts


def _disk_section_points(
    radius=10.0, theta=-90, start_angle=0, angle_resolution=2.5, layer=LAYER.WG
):
    angle1 = (start_angle) * pi / 180
    angle2 = (start_angle + theta) * pi / 180
    t = np.linspace(angle1, angle2, int(abs(theta) / angle_resolution))
    xpts = (radius * cos(t)).tolist()
    ypts = (radius * sin(t)).tolist()
    xpts.append(0)
    ypts.append(0)
    return xpts, ypts


@pp.autoname
def bend_circular(
    radius: float = 10.0,
    width: float = 0.5,
    theta: int = -90,
    start_angle: int = 0,
    angle_resolution: float = 2.5,
    layer: Tuple[int, int] = LAYER.WG,
    layers_cladding: List[Tuple[int, int]] = [pp.LAYER.WGCLAD],
    cladding_offset: float = 3.0,
) -> Component:
    """ Creates an arc of arclength ``theta`` starting at angle ``start_angle``

    Args:
        radius
        width: of the waveguide
        theta: arc length
        start_angle:
        angle_resolution
        layer

    .. plot::
      :include-source:

      import pp

      c = pp.c.bend_circular(
        radius=10,
        width=0.5,
        theta=-90,
        start_angle=0,
      )
      pp.plotgds(c)

    """
    component = pp.Component()

    # Core
    inner_radius = radius - width / 2
    outer_radius = radius + width / 2
    angle1 = (start_angle) * pi / 180
    angle2 = (start_angle + theta) * pi / 180
    t = np.linspace(angle1, angle2, int(abs(theta) / angle_resolution))
    inner_points_x = (inner_radius * cos(t)).tolist()
    inner_points_y = (inner_radius * sin(t)).tolist()
    outer_points_x = (outer_radius * cos(t)).tolist()
    outer_points_y = (outer_radius * sin(t)).tolist()
    xpts = inner_points_x + outer_points_x[::-1]
    ypts = inner_points_y + outer_points_y[::-1]

    component.add_polygon(points=(xpts, ypts), layer=layer)

    # Cladding
    w = width + 2 * cladding_offset
    inner_radius = radius - w / 2
    outer_radius = radius + w / 2
    angle1 = (start_angle) * pi / 180
    angle2 = (start_angle + theta) * pi / 180
    t = np.linspace(angle1, angle2, int(abs(theta) / angle_resolution))
    inner_points_x = (inner_radius * cos(t)).tolist()
    inner_points_y = (inner_radius * sin(t)).tolist()
    outer_points_x = (outer_radius * cos(t)).tolist()
    outer_points_y = (outer_radius * sin(t)).tolist()
    xpts = inner_points_x + outer_points_x[::-1]
    ypts = inner_points_y + outer_points_y[::-1]

    for layer_cladding in layers_cladding:
        component.add_polygon(points=(xpts, ypts), layer=layer_cladding)

    midpoint1 = (radius * cos(angle1), radius * sin(angle1))
    component.add_port(
        name="W0",
        midpoint=midpoint1,
        width=width,
        orientation=start_angle - 90 + 180 * (theta < 0),
        layer=layer,
    )
    midpoint2 = (radius * cos(angle2), radius * sin(angle2))
    component.add_port(
        name="N0",
        midpoint=midpoint2,
        width=width,
        orientation=start_angle + theta + 90 - 180 * (theta < 0),
        layer=layer,
    )
    component.info["length"] = (abs(theta) * pi / 180) * radius
    component.radius = radius
    component.width = width
    component.move((0, radius))

    assert pp.drc.on_grid(
        midpoint1[1] - width / 2
    ), f"x_input point is off grid {midpoint1[1] - width/2}"
    assert pp.drc.on_grid(
        midpoint2[0] - width / 2
    ), f"y_output popint is off grid {midpoint1[1] - width/2}"

    pp.port.rename_ports_by_orientation(component)
    return component


@pp.autoname
def bend_circular_deep_rib(layer=pp.LAYER.SLAB90, layers_cladding=[], **kwargs):
    c = bend_circular(layer=layer, layers_cladding=layers_cladding, **kwargs)
    pp.port.rename_ports_by_orientation(c)
    return c


@pp.autoname
def bend_circular_shallow_rib(layer=pp.LAYER.SLAB150, layers_cladding=[], **kwargs):
    return bend_circular(layer=layer, layers_cladding=layers_cladding, **kwargs)


@pp.autoname
def _bend_circular(
    radius=10.0,
    width=0.5,
    theta=-90,
    start_angle=0,
    angle_resolution=2.5,
    layer=LAYER.WG,
):

    component = pp.Component()

    xpts, ypts = _bend_points(radius, width, theta, start_angle, angle_resolution)
    angle1 = (start_angle) * pi / 180
    angle2 = (start_angle + theta) * pi / 180

    component.add_polygon(points=(xpts, ypts), layer=layer)

    component.add_port(
        name="W0",
        midpoint=(radius * cos(angle1), radius * sin(angle1)),
        width=width,
        orientation=start_angle - 90 + 180 * (theta < 0),
        layer=layer,
    )
    component.add_port(
        name="N0",
        midpoint=(radius * cos(angle2), radius * sin(angle2)),
        width=width,
        orientation=start_angle + theta + 90 - 180 * (theta < 0),
        layer=layer,
    )
    component.info["length"] = (abs(theta) * pi / 180) * radius
    component.radius = radius
    component.width = width
    component.move((0, radius))

    pp.port.auto_rename_ports(component)
    return component


@pp.autoname
def bend_circular180(
    radius: Union[int, float] = 10.0,
    width: float = 0.5,
    theta: int = 180,
    start_angle: int = -90,
    angle_resolution: float = 2.5,
    layer: Tuple[int, int] = LAYER.WG,
    **kwargs,
) -> Component:
    c = bend_circular(
        radius=radius,
        width=width,
        theta=theta,
        start_angle=start_angle,
        angle_resolution=angle_resolution,
        layer=layer,
        **kwargs,
    )
    return c


def _bend_circular_windows(
    radius=10,
    start_angle=0,
    theta=-90,
    angle_resolution=2.5,
    windows=[-0.25, 0.25, LAYER.WG],
):
    """
    windows: [(y_start, y_stop, layer), ...]
    """
    component = pp.Component()
    y_min, y_max, layer0 = windows[0]
    y_min, y_max = min(y_min, y_max), max(y_min, y_max)

    bend_params = {
        "start_angle": start_angle,
        "theta": theta,
        "angle_resolution": angle_resolution,
    }

    # Create each bend shape
    for y_start, y_stop, layer in windows:
        w = abs(y_stop - y_start)
        y = (y_stop + y_start) / 2
        r = radius - y
        _c = bend_circular(radius=r, width=w, layer=layer, **bend_params).ref()
        _c.movey(y)
        component.add(_c)
        component.absorb(_c)
        y_min = min(y_stop, y_start, y_min)
        y_max = max(y_stop, y_start, y_max)
    width = y_max - y_min

    # Define ports
    angle1 = (start_angle) * pi / 180
    angle2 = (start_angle + theta) * pi / 180

    component.add_port(
        name="tmp0",
        midpoint=(radius * cos(angle1), radius + radius * sin(angle1)),
        width=width,
        orientation=start_angle - 90 + 180 * (theta < 0),
        layer=layer0,
    )
    component.add_port(
        name="tmp1",
        midpoint=(radius * cos(angle2), radius + radius * sin(angle2)),
        width=width,
        orientation=start_angle + theta + 90 - 180 * (theta < 0),
        layer=layer0,
    )
    pp.port.auto_rename_ports(component)
    return component


@pp.autoname
def bend_circular_trenches(
    width=0.5, trench_width=3.0, trench_offset=0.2, trench_layer=LAYER.SLAB90, **kwargs
):
    """ defines trenches """
    w = width / 2
    ww = w + trench_width
    wt = ww + trench_offset
    windows = [(-ww, ww, LAYER.WG), (-wt, -w, trench_layer), (w, wt, trench_layer)]
    return _bend_circular_windows(windows=windows, **kwargs)


@pp.autoname
def bend_circular_slab(width=0.5, cladding=2.0, slab_layer=LAYER.SLAB90, **kwargs):
    a = width / 2
    b = a + cladding
    windows = [(-a, a, LAYER.WG), (-b, b, slab_layer)]
    return _bend_circular_windows(windows=windows, **kwargs)


bend_circular_ridge = bend_circular_trenches


@pp.autoname
def bend_circular_slot(width=0.5, gap=0.2, **kwargs):
    a = width / 2
    d = a + gap / 2
    windows = [(-a - d, a - d, LAYER.WG), (-a + d, a + d, LAYER.WG)]
    return _bend_circular_windows(windows=windows, **kwargs)


def _demo_bend():
    c = bend_circular()
    pp.write_gds(c)


if __name__ == "__main__":
    # c = bend_circular_trenches()
    # c = bend_circular_deep_rib()
    # print(c.ports)
    # c = bend_circular(radius=5.0005, width=1.002, theta=180, pins=True)
    c = bend_circular(theta=180, pins=True)
    print(c.ports.keys())
    # print(c.ports["N0"].midpoint)
    # print(c.settings)
    # c = bend_circular_slot()
    # c = bend_circular(width=0.45, radius=5)
    # print(c.ports)
    pp.show(c)
    pp.plotgds(c)



import copy
from pp.component import Component, ComponentReference, Port
from pp.config import call_if_func
from pp.name import autoname
from phidl.device_layout import Device


@autoname
def import_phidl_component(component: Device, **kwargs) -> Component:
    """ returns a gdsfactory Component from a phidl Device or function
    """
    D = call_if_func(component, **kwargs)
    D_copy = Component(name=D._internal_name)
    D_copy.info = copy.deepcopy(D.info)
    for ref in D.references:
        new_ref = ComponentReference(
            device=ref.parent,
            origin=ref.origin,
            rotation=ref.rotation,
            magnification=ref.magnification,
            x_reflection=ref.x_reflection,
        )
        new_ref.owner = D_copy
        D_copy.add(new_ref)
        for alias_name, alias_ref in D.aliases.items():
            if alias_ref == ref:
                D_copy.aliases[alias_name] = new_ref

    for p in D.ports.values():
        D_copy.add_port(
            port=Port(
                name=p.name,
                midpoint=p.midpoint,
                width=p.width,
                orientation=p.orientation,
                parent=p.parent,
            )
        )
    for poly in D.polygons:
        D_copy.add_polygon(poly)
    for label in D.labels:
        D_copy.add_label(
            text=label.text,
            position=label.position,
            layer=(label.layer, label.texttype),
        )
    return D_copy


if __name__ == "__main__":
    import pp
    import phidl.geometry as pg

    c = pg.rectangle()
    c = pg.snspd()

    c2 = import_phidl_component(component=c)
    print(c2.ports)
    pp.show(c2)



wm = 1e-9
wm_min = 0.9e-9  # 1nm
wm_max = 999e-9  # 1um

wum = wm * 1e6
wum_min = wm_min * 1e6
wum_max = wm_max * 1e6


def assert_um(w):
    if hasattr(w, "__iter__"):
        w = w[0]
    assert wum_min < w < wum_max, f"are you sure `{w}` is in um?"


def assert_m(w):
    if hasattr(w, "__iter__"):
        w = w[0]
    assert wm_min < w < wm_max, f"are you sure `{w}` is in m?"


def assert_dbu(value, dbu=1e-3):
    if hasattr(value, "__iter__"):
        value = value[0]
    assert (
        value > dbu
    ), f"make sure {value} is in um and is larger than the minimum database unit {dbu}"


if __name__ == "__main__":
    w = [1e-9, 2e-9]
    assert_m(w)
    w = [1.55]
    assert_m(w)



""" based on phidl tutorial
"""

import pp


# ==============================================================================
# We'll start by assuming we have a function waveguide() which already exists
# and makes us a simple waveguide rectangle.  Many functions like this
# exist in the pp.components library and are ready-for-use.  We write this
# one out fully just so it's explicitly clear what's happening
# ==============================================================================


def waveguide_sample(length=5, width=1):
    wg = pp.Component("waveguide_sample")
    wg.add_polygon([(0, 0), (length, 0), (length, width), (0, width)], layer=(2, 0))
    wg.add_port(name="W0", midpoint=[0, width / 2], width=width, orientation=180)
    wg.add_port(name="E0", midpoint=[length, width / 2], width=width, orientation=0)
    return wg


# ==============================================================================
# Create a blank component
# ==============================================================================
# Let's create a new Component ``c`` which will act as a blank canvas (c can be
# thought of as a blank GDS cell with some special features). Note that when we
# make a Component
c = pp.Component("MultiWaveguide")

# Now say we want to add a few waveguides to to our  Component" c.
# First we create the waveguides.  As you can see from the waveguide() function
# definition, the waveguide() function creates another Component ("WG").
# This can be thought of as the waveguide() function creating another GDS cell,
# only this one has some geometry inside it.
#
# Let's create two of these Devices by calling the waveguide() function
WG1 = waveguide_sample(length=10, width=1)
WG2 = waveguide_sample(length=12, width=2)

# Now we've made two waveguides Component WG1 and WG2, and we have a blank
# Component c. We can add references from the devices WG1 and WG2 to our blank
# Component byz using the add_ref() function.
# After adding WG1, we see that the add_ref() function returns a handle to our
# reference, which we will label with lowercase letters wg1 and wg2.  This
# handle will be useful later when we want to move wg1 and wg2 around in c.
wg1 = c.add_ref(WG1)  # Using the function add_ref()
wg2 = c << WG2  # Using the << operator which is identical to add_ref()

# Alternatively, we can do this all on one line
wg3 = c.add_ref(waveguide_sample(length=14, width=3))

pp.qp(c)  # quickplot it!
pp.show(c)  # show it in Klayout



from omegaconf import OmegaConf


def test_netlist_write():
    from pp.components.mzi import mzi

    c = mzi()
    netlist = c.get_netlist()
    # netlist.pop('connections')
    OmegaConf.save(netlist, "mzi.yml")


if __name__ == "__main__":
    c = test_netlist_write()



""" adiabatic tapers from CSV files
"""
import pathlib

import pp

data_path = pathlib.Path(__file__).parent / "csv_data"


def taper_from_csv(csv_path, wg_layer=1, clad_offset=3, clad_layer=pp.LAYER.WGCLAD):
    taper_data = pp.load_csv(csv_path)
    # taper_data = pd.read_csv(csv_path)
    # print(taper_data)
    xs = taper_data["x"] * 1e6
    ys = taper_data["width"] * 1e6 / 2.0
    ys_trench = ys + clad_offset

    c = pp.Component()
    c.add_polygon(list(zip(xs, ys)) + list(zip(xs, -ys))[::-1], layer=wg_layer)
    c.add_polygon(
        list(zip(xs, ys_trench)) + list(zip(xs, -ys_trench))[::-1], layer=clad_layer
    )

    c.add_port(
        name="W0",
        midpoint=(xs[0], 0),
        width=2 * ys[0],
        orientation=180,
        port_type="optical",
    )
    c.add_port(
        name="E0",
        midpoint=(xs[-1], 0),
        width=2 * ys[-1],
        orientation=0,
        port_type="optical",
    )
    return c


@pp.autoname
def taper_w3_l36(**kwargs):
    csv_path = data_path / "taper_strip_0p5_3_36.csv"
    return taper_from_csv(csv_path, **kwargs)


@pp.autoname
def taper_w10_l100(**kwargs):
    csv_path = data_path / "taper_strip_0p5_10_100.csv"
    return taper_from_csv(csv_path, **kwargs)


@pp.autoname
def taper_w10_l150(**kwargs):
    csv_path = data_path / "taper_strip_0p5_10_150.csv"
    return taper_from_csv(csv_path, **kwargs)


@pp.autoname
def taper_w10_l200(**kwargs):
    csv_path = data_path / "taper_strip_0p5_10_200.csv"
    return taper_from_csv(csv_path, **kwargs)


@pp.autoname
def taper_w11_l200(**kwargs):
    csv_path = data_path / "taper_strip_0p5_11_200.csv"
    return taper_from_csv(csv_path, **kwargs)


@pp.autoname
def taper_w12_l200(**kwargs):
    csv_path = data_path / "taper_strip_0p5_12_200.csv"
    return taper_from_csv(csv_path, **kwargs)


if __name__ == "__main__":
    c = taper_w3_l36()
    # c = taper_w10_l100()
    # c = taper_w11_l200()
    pp.show(c)



from typing import Any, Callable, List, Optional, Tuple, Union
from numpy import float64
from phidl.device_layout import Label
from pp.component import Component, ComponentReference
from pp.layers import LAYER
from pp.components.bend_circular import bend_circular
from pp.components import waveguide
from pp.components.grating_coupler.elliptical_trenches import grating_coupler_te

from pp.routing.manhattan import round_corners
from pp.routing.connect_bundle import link_optical_ports
from pp.routing.connect_bundle import get_min_spacing
from pp.routing.route_south import route_south

from pp.routing.utils import direction_ports_from_list_ports

from pp.routing.connect import connect_strip_way_points
from pp.routing.connect import get_waypoints_connect_strip
from pp.routing.get_input_labels import get_input_labels
from pp.port import select_optical_ports
import pp

SPACING_GC = 127.0
BEND_RADIUS = pp.conf.tech.bend_radius


def route_fiber_array(
    component: Component,
    optical_io_spacing: float = SPACING_GC,
    grating_coupler: Callable = grating_coupler_te,
    bend_factory: Callable = bend_circular,
    straight_factory: Callable = waveguide,
    fanout_length: Optional[int] = None,
    max_y0_optical: None = None,
    with_align_ports: bool = True,
    waveguide_separation: float = 4.0,
    optical_routing_type: Optional[int] = None,
    bend_radius: float = BEND_RADIUS,
    connected_port_list_ids: None = None,
    nb_optical_ports_lines: int = 1,
    force_manhattan: bool = False,
    excluded_ports: List[Any] = [],
    grating_indices: None = None,
    route_filter: Callable = connect_strip_way_points,
    gc_port_name: str = "W0",
    gc_rotation: int = -90,
    layer_label: Tuple[int, int] = LAYER.LABEL,
    component_name: Optional[str] = None,
    x_grating_offset: int = 0,
    optical_port_labels: None = None,
    route_factory: Callable = route_south,
    get_input_labels_function: Callable = get_input_labels,
    select_ports: Callable = select_optical_ports,
) -> Tuple[
    List[Union[ComponentReference, Label]], List[List[ComponentReference]], float64
]:
    """
    Returns component I/O elements for adding grating couplers with a fiber array input
    Many components are fine with the default.

    Args:
        component: The component to connect.
        optical_io_spacing: the wanted spacing between the optical I/O
        grating_coupler: grating coupler instance, function or list of functions
        bend_factory: bend_circular
        straight_factory: waveguide
        fanout_length: Wanted distance between the gratings and the southest component port. If None, automatically calculated.
        max_y0_optical: Maximum y coordinate at which the intermediate optical ports can be set. Usually fine to leave at None.
        with_align_ports: If True, add compact loopback alignment ports
        waveguide_separation: min separation between the waveguides used to route grating couplers to the component I/O.
        optical_routing_type: There are three options for optical routing
           * ``0`` is very basic but can be more compact.  Can also be used in combination with ``connected_port_list_ids`` or to route some components which otherwise fail with type ``1``.
           * ``1`` is the standard routing.
           * ``2`` uses the optical ports as a guideline for the component's physical size (instead of using the actual component size).  Useful where the component is large due to metal tracks
           * ``None: leads to an automatic decision based on size and number
           of I/O of the component.
        bend_radius: bend radius
        connected_port_list_ids: only for type 0 optical routing.  Can specify which ports goes to which grating assuming the gratings are ordered from left to right.  e.g ['N0', 'W1','W0','E0','E1', 'N1' ] or [4,1,7,3]
        nb_optical_ports_lines: number of lines with I/O grating couplers.  One line by default.  WARNING: Only works properly if:
            - nb_optical_ports_lines divides the total number of ports
            - the components have an equal number of inputs and outputs
        force_manhattan: in some instances, the port linker defaults to an S-bend due to lack of space to do manhattan. Force manhattan offsets all the ports to replace the s-bend by a straight link.  This fails if multiple ports have the same issue
        excluded_ports: ports excluded from routing
        grating_indices: allows to fine skip some grating slots e.g [0,1,4,5] will put two gratings separated by the pitch. Then there will be two empty grating slots, and after that an additional two gratings.
        route_filter: waveguide and bend factories
        gc_port_name: grating_coupler port name, where to route waveguides
        gc_rotation: grating_coupler rotation (deg)
        layer_label: for TM labels
        component_name: name of component
        x_grating_offset: x offset
        optical_port_labels: port labels that need connection
        route_factory: factories for route
        get_input_labels_function: functions to add labels
        select_ports: function to select ports

    Returns:
        elements, io_grating_lines, y0_optical
    """
    if optical_port_labels is None:
        # for pn, p in component.ports.items():
        #     print(p.name, p.port_type, p.layer)
        # optical_ports = component.get_ports_list(port_type='optical')
        optical_ports = list(select_ports(component.ports).values())
        # print(optical_ports)
    else:
        optical_ports = [component.ports[lbl] for lbl in optical_port_labels]

    optical_ports = [p for p in optical_ports if p.name not in excluded_ports]
    N = len(optical_ports)
    if N == 0:
        return [], [], 0

    elements = []

    """
    # grating_coupler can either be a gratings/factories or a list of  gratings/factories
    """

    if isinstance(grating_coupler, list):
        grating_couplers = [pp.call_if_func(g) for g in grating_coupler]
        grating_coupler = grating_couplers[0]
    else:
        grating_coupler = pp.call_if_func(grating_coupler)
        grating_couplers = [grating_coupler] * N

    assert (
        gc_port_name in grating_coupler.ports
    ), f"{gc_port_name} not in {list(grating_coupler.ports.keys())}"

    """
    # Now:
    # - grating_coupler is a single grating coupler
    # - grating_couplers is a list of grating couplers
    """

    """
    # Define the route filter to apply to connection methods
    """

    route_filter_params = {
        "bend_radius": bend_radius,
        "wg_width": grating_coupler.ports[gc_port_name].width,
    }

    def routing_method(p1, p2, **kwargs):
        way_points = get_waypoints_connect_strip(p1, p2, **kwargs)
        return route_filter(way_points, **route_filter_params)

    R = bend_radius

    """
    # `delta_gr_min` Used to avoid crossing between waveguides in special cases
    # This could happen when abs(x_port - x_grating) <= 2 * bend_radius
    """

    delta_gr_min = 2 * bend_radius + 1

    io_sep = optical_io_spacing
    offset = (N - 1) * io_sep / 2.0

    # Get the center along x axis
    x_c = round(sum([p.x for p in optical_ports]) / N, 1)
    y_min = component.ymin  # min([p.y for p in optical_ports])

    # Sort the list of optical ports:
    direction_ports = direction_ports_from_list_ports(optical_ports)
    sep = waveguide_separation

    K = len(optical_ports)
    K = K + 1 if K % 2 else K

    # Set routing type if not specified
    pxs = [p.x for p in optical_ports]
    is_big_component = (
        (K > 2)
        or (max(pxs) - min(pxs) > io_sep - delta_gr_min)
        or (component.xsize > io_sep)
    )
    if optical_routing_type is None:
        if not is_big_component:
            optical_routing_type = 0
        else:
            optical_routing_type = 1
    """
    Look at a bunch of conditions to choose the default length
    if the default fanout distance is not set
    """

    def has_p(side):
        return len(direction_ports[side]) > 0

    list_ew_ports_on_sides = [has_p(side) for side in ["E", "W"]]
    list_ns_ports_on_sides = [has_p(side) for side in ["N", "S"]]

    has_ew_ports = any(list_ew_ports_on_sides)
    has_ns_ports = any(list_ns_ports_on_sides)

    is_one_sided_horizontal = False
    for side1, side2 in [("E", "W"), ("W", "E")]:
        if len(direction_ports[side1]) >= 2:
            if all([len(direction_ports[side]) == 0 for side in ["N", "S", side2]]):
                is_one_sided_horizontal = True

    # Compute fanout length if not specified
    if fanout_length is None:
        fanout_length = bend_radius + 1.0
        # We need 3 bends in that case to connect the most bottom port to the
        # grating couplers
        if has_ew_ports and is_big_component:
            fanout_length = max(fanout_length, 3 * bend_radius + 1.0)

        if has_ns_ports or is_one_sided_horizontal:
            fanout_length = max(fanout_length, 2 * bend_radius + 1.0)

        if has_ew_ports and not is_big_component:
            fanout_length = max(fanout_length, bend_radius + 1.0)

    fanout_length += 5
    # use x for grating coupler since we rotate it
    y0_optical = y_min - fanout_length - grating_coupler.ports[gc_port_name].x
    y0_optical += -K / 2 * sep
    y0_optical = round(y0_optical, 1)

    if max_y0_optical is not None:
        y0_optical = round(min(max_y0_optical, y0_optical), 1)
    """
     - First connect half of the north ports going from middle of list
    down to first elements
     - then connect west ports (top to bottom)
     - then connect south ports (left to right)
     - then east ports (bottom to top)
     - then second half of the north ports (right to left)

    """
    north_ports = direction_ports["N"]
    north_start = north_ports[0 : len(north_ports) // 2]
    north_finish = north_ports[len(north_ports) // 2 :]

    west_ports = direction_ports["W"]
    west_ports.reverse()
    east_ports = direction_ports["E"]
    south_ports = direction_ports["S"]
    north_finish.reverse()  # Sort right to left
    north_start.reverse()  # Sort right to left
    ordered_ports = north_start + west_ports + south_ports + east_ports + north_finish

    nb_ports_per_line = N // nb_optical_ports_lines
    grating_coupler_si = grating_coupler.size_info
    y_gr_gap = (K / (nb_optical_ports_lines) + 1) * sep
    gr_coupler_y_sep = grating_coupler_si.height + y_gr_gap + bend_radius

    offset = (nb_ports_per_line - 1) * io_sep / 2 - x_grating_offset
    io_gratings_lines = []  # [[gr11, gr12, gr13...], [gr21, gr22, gr23...] ...]

    if grating_indices is None:
        grating_indices = list(range(nb_ports_per_line))
    else:
        assert len(grating_indices) == nb_ports_per_line

    for j in range(nb_optical_ports_lines):
        io_gratings = [
            gc.ref(
                position=(x_c - offset + i * io_sep, y0_optical - j * gr_coupler_y_sep),
                rotation=gc_rotation,
                port_id=gc_port_name,
            )
            for i, gc in zip(grating_indices, grating_couplers)
        ]

        io_gratings_lines += [io_gratings[:]]

    # optical routing - type ``0``
    if optical_routing_type == 0:
        """
        Basic optical routing, typically fine for small components
        No heuristic to avoid collisions between connectors.

        If specified ports to connect in a specific order
        (i.e if connected_port_list_ids is not None and not empty)
        then grab these ports
        """

        if connected_port_list_ids:
            ordered_ports = [component.ports[i] for i in connected_port_list_ids]

        for io_gratings in io_gratings_lines:
            for i in range(N):
                p0 = io_gratings[i].ports[gc_port_name]
                p1 = ordered_ports[i]
                elements += [routing_method(p0, p1, bend_radius=bend_radius)]

    # optical routing - type ``1 or 2``
    elif optical_routing_type in [1, 2]:
        elems, to_route = route_factory(
            component=component,
            bend_radius=bend_radius,
            optical_routing_type=optical_routing_type,
            excluded_ports=excluded_ports,
            waveguide_separation=waveguide_separation,
            io_gratings_lines=io_gratings_lines,
            gc_port_name=gc_port_name,
            route_filter=route_filter,
        )
        elements += elems

        if force_manhattan:
            """
            1) find the min x_distance between each grating port and
            each component port.

            2) If abs(min distance) < 2* bend radius
                then offset io_gratings by -min_distance

            """
            min_dist = 2 * R + 10.0
            min_dist_threshold = 2 * R + 1.0
            for io_gratings in io_gratings_lines:
                for gr in io_gratings:
                    for p in to_route:
                        dist = gr[0].x - p.x
                        if abs(dist) < abs(min_dist):
                            min_dist = dist
                if abs(min_dist) < min_dist_threshold:
                    for gr in io_gratings:
                        gr.translate((-min_dist, 0))

        # If the array of gratings is too close, adjust its location
        gc_ports_tmp = []
        for io_gratings in io_gratings_lines:
            gc_ports_tmp += [gc.ports[gc_port_name] for gc in io_gratings]
        min_y = get_min_spacing(to_route, gc_ports_tmp, sep=sep, radius=R)
        delta_y = abs(to_route[0].y - gc_ports_tmp[0].y)

        if min_y > delta_y:
            for io_gratings in io_gratings_lines:
                for gr in io_gratings:
                    gr.translate(0, delta_y - min_y)

        # If we add align ports, we need enough space for the bends
        end_straight_offset = waveguide_separation + 5 if with_align_ports else 0.01
        if len(io_gratings_lines) == 1:
            io_gratings = io_gratings_lines[0]
            gc_ports = [gc.ports[gc_port_name] for gc in io_gratings]
            elements += link_optical_ports(
                to_route,
                gc_ports,
                separation=sep,
                end_straight_offset=end_straight_offset,
                route_filter=route_filter,
                **route_filter_params,
            )

        else:
            for io_gratings in io_gratings_lines:
                gc_ports = [gc.ports[gc_port_name] for gc in io_gratings]
                nb_gc_ports = len(io_gratings)
                nb_ports_to_route = len(to_route)
                n0 = nb_ports_to_route / 2
                dn = nb_gc_ports / 2
                elements += link_optical_ports(
                    to_route[n0 - dn : n0 + dn],
                    gc_ports,
                    separation=sep,
                    end_straight_offset=end_straight_offset,
                    route_filter=route_filter,
                    **route_filter_params,
                )
                del to_route[n0 - dn : n0 + dn]

    if with_align_ports:
        """
        Add loop back with alignment ports
        """
        gca1, gca2 = [
            grating_coupler.ref(
                position=(
                    x_c - offset + ii * io_sep,
                    io_gratings_lines[-1][0].ports[gc_port_name].y,
                ),
                rotation=gc_rotation,
                port_id=gc_port_name,
            )
            for ii in [grating_indices[0] - 1, grating_indices[-1] + 1]
        ]

        gsi = grating_coupler.size_info
        p0 = gca1.ports[gc_port_name].position
        p1 = gca2.ports[gc_port_name].position
        bend_radius_align_ports = R
        a = bend_radius_align_ports + 5.0  # 0.5
        b = max(2 * a, io_sep / 2)
        y_bot_align_route = -gsi.width - waveguide_separation

        route = [
            p0,
            p0 + (0, a),
            p0 + (b, a),
            p0 + (b, y_bot_align_route),
            p1 + (-b, y_bot_align_route),
            p1 + (-b, a),
            p1 + (0, a),
            p1,
        ]
        elements += [gca1, gca2]

        bend90 = bend_factory(radius=bend_radius)
        loop_back = round_corners(route, bend90, straight_factory)
        elements += [loop_back]

    """ input_label for automated testing opt_TE_1550_componentName_0_portLabel"""
    elements += get_input_labels_function(
        io_gratings, ordered_ports, component_name, layer_label, gc_port_name
    )

    return elements, io_gratings_lines, y0_optical


if __name__ == "__main__":
    gcte = pp.c.grating_coupler_te
    gctm = pp.c.grating_coupler_tm

    c = pp.c.mmi2x2()
    c = pp.c.waveguide(length=500)
    c.ports = c.get_ports_dict(prefix="W")

    elements, gc, _ = route_fiber_array(c, grating_coupler=[gcte, gctm, gcte, gctm])
    for e in elements:
        c.add(e)
    for e in gc:
        c.add(e)
    pp.show(c)



from typing import Dict, List, Tuple, Union
import numpy as np
from numpy import ndarray

from pp.component import Component
from pp.component import ComponentReference
from pp.port import select_electrical_ports

IDENTITY = (0, False)
R90 = (90, False)
R180 = (180, False)
R270 = (270, False)
mirror_x = (0, True)
mirror_y = (180, True)
mirror_x270 = (270, True)
mirror_x90 = (90, True)

TRANSFORMATION_MAP = {
    IDENTITY: "None",
    R90: "R90",
    R180: "R180",
    R270: "R270",
    mirror_y: "mirror_y",
    mirror_x: "mirror_x",
    mirror_x270: "mirror_x270",
    mirror_x90: "mirror_x90",
}

STR_TO_TRANSFORMATION_MAP = {v: k for k, v in TRANSFORMATION_MAP.items()}


def get_elec_ports_from_component_names(component, names=[]):
    """
    Args
        component <pp.Component>; should have component.info["instances"]
    """
    e_ports = {}

    for name in names:
        _ports = {
            "{}_{}".format(name, p.name): p
            for p in select_electrical_ports(component.info["instances"][name]).values()
        }
        e_ports.update(_ports)

    # update port names
    for pname, p in e_ports.items():
        p.name = pname

    return e_ports


def gen_sref(
    component: Component,
    transformation_name: str,
    port_name: str,
    position: Union[Tuple[int, int], ndarray],
) -> ComponentReference:
    """"""

    if transformation_name not in TRANSFORMATION_MAP.values():
        raise ValueError(
            "{} is not a valid transformation({})".format(
                transformation_name, list(TRANSFORMATION_MAP.values())
            )
        )
    rotation_angle, x_reflection = STR_TO_TRANSFORMATION_MAP[transformation_name]
    position = np.array(position)

    if port_name is None:
        port_position = np.array([0, 0])
    else:
        if port_name not in component.ports.keys():
            raise ValueError(
                "{} port name not in {} ports ({})".format(
                    port_name, component.name, component.ports.keys()
                )
            )
        port_position = component.ports[port_name].midpoint

    ref = ComponentReference(component=component, origin=(0, 0))

    if x_reflection:  # Vertical mirror: Reflection across x-axis
        y0 = port_position[1]
        ref.reflect(p1=(0, y0), p2=(1, y0))

    if rotation_angle != 0:
        ref.rotate(rotation_angle, center=port_position)

    translation = np.array(position - port_position)

    ref.move(destination=translation)

    return ref


def netlist_to_component(
    instances: Dict[str, Tuple[Component, str]],
    connections: List[Tuple[str, str, str, str]],
    ports_map: Dict[str, Tuple[str, str]] = None,
    position: Tuple[float, float] = (0.0, 0.0),
) -> Component:
    """ Netlist_to_component is deprecated! use pp.componet_from_yaml instead
    Returns a component from a netlist (instances, connections and ports map)

    Args:
        instances:
            list of (instance_id <str>, component <Component>, transform <tuple>)

        connections:
            list of (component_id1, port_name1, component_id2, port_name2)
            Has to be ordered such that any component refered to in the connections
            has already been placed (except for the first component)

        ports_map:
            {port_name: (instance_id, port_name)}

    Returns: component with netlist stored in component.netlist

    [
        {
            "name": "CP2x2",
            "rank": 0,
            "ports": ["in1", "in2", "out1", "out2"],
            "type": "CPBIDIR",
            "settings": {"R": 0.5},
        },
        {
            "name": "WG_CAVITY",
            "rank": 0,
            "ports": ["in", "out"],
            "type": "WG",
            "settings": {"length": 50, "loss_dB_m": 60000, "width": 0.5},
        },
        {
            "name": "ring_resonator",
            "rank": 1,
            "type": "COMPOUND",
            "settings": {},
            "instances": [
                "CP1, CP2x2, None, None, 0.0, 0.0",
                "WG1, WG_CAVITY, None, None, 0.0, 0.0",
            ],
            "connections": ["CP1, out1, WG1, in", "WG1, out, CP1, in1"],
            "ports": {"in1": "CP1, in2", "out1": "CP1, out2"},
        },
    ]

    mirror, rotation, x, y
    """

    if len(connections) == 0:
        raise ValueError("no connections defined")

    instance_id, port, _, _ = connections[0]
    assert instance_id in instances, f"{instance_id} not in {list(instances.keys())}"
    component, transform_name = instances[instance_id]

    # First component reference
    sref_start = gen_sref(component, transform_name, port, position)
    cmp_name_to_sref = {instance_id: sref_start}

    # Iterate over all connections: create and place instances
    for cmp1_name, port1, cmp2_name, port2 in connections:
        if cmp1_name not in cmp_name_to_sref:
            cmp1_name, port1, cmp2_name, port2 = cmp2_name, port2, cmp1_name, port1

        if cmp2_name not in cmp_name_to_sref:
            component, transform_name = instances[cmp2_name]

            _ref = cmp_name_to_sref[cmp1_name]
            try:
                position = _ref.ports[port1].midpoint
            except Exception:
                print("{} has not port {}".format(cmp1_name, port1))

            sref = gen_sref(component, transform_name, port2, position)

            cmp_name_to_sref[cmp2_name] = sref

    c = Component()
    c.add(list(cmp_name_to_sref.values()))

    if ports_map:
        for port_name, (cmp_id, internal_port_name) in ports_map.items():
            component_ref = cmp_name_to_sref[cmp_id]
            component = component_ref.parent
            assert internal_port_name in component.ports, (
                f"{internal_port_name} not in {component_ref.ports.keys()} for"
                f" {component}"
            )
            port = component_ref.ports[internal_port_name]
            c.add_port(port_name, port=port)
            ports_map = {k: ", ".join(v) for k, v in ports_map.items()}

    # Set aliases
    for cmp_id, ref in cmp_name_to_sref.items():
        c[cmp_id] = ref

    c.info["instances"] = cmp_name_to_sref

    # add leaf cells to netlist
    netlist = []
    for name, (component, _) in instances.items():
        netlist.append(
            dict(
                name=name,
                rank=0,
                ports=list(component.ports.keys()),
                settings=component.settings,
            )
        )
    # add the compound cell to the netlist
    netlist.append(
        dict(
            name="component_name",
            rank=1,
            type="COMPOUND",
            settings={},
            connections=[", ".join(i) for i in connections],
            ports=ports_map,
        )
    )
    c.netlist = netlist
    return c


def test_netlist_ring():
    from pp.components.ring_single_bus import ring_single_bus_netlist

    instances, connections, ports_map = ring_single_bus_netlist()
    c = netlist_to_component(instances, connections, ports_map)
    # print((c.get_dependencies()))
    # print(len(c.get_dependencies()))
    assert len(c.get_dependencies()) == 4
    return c


if __name__ == "__main__":
    import pp

    # print(c.netlist)
    c = test_netlist_ring()
    pp.show(c)



from typing import Callable
from pp.components import bend_circular
from pp.components import wg_heater_connected as waveguide_heater
from pp.components import waveguide
from pp.components import coupler
from pp.netlist_to_gds import netlist_to_component
from pp.name import autoname
from pp.routing import route_elec_ports_to_side
from pp.port import select_electrical_ports

from pp.components.extension import line
from pp.components.component_sequence import component_sequence
from pp.component import Component


@autoname
def mzi_arm(
    L0: float = 60.0,
    DL: float = 0.0,
    L_top: float = 10.0,
    bend_radius: float = 10.0,
    bend90_factory: Callable = bend_circular,
    straight_heater_factory: Callable = waveguide_heater,
    straight_factory: Callable = waveguide,
    with_elec_connections: bool = True,
) -> Component:
    """

    Args:
        L0: vertical length with heater
        DL: extra vertical length without heater (delat_length=2*DL)
        L_top: 10.0, horizontal length
        bend_radius: 10.0
        bend90_factory: bend_circular
        straight_heater_factory: waveguide_heater
        straight_factory: waveguide

    ::

         L_top
        |     |
        DL    DL
        |     |
        L0    L0
        |     |
       -|     |-

        B2-Sh1-B3
        |     |
        Sv1   Sv2
        |     |
        H1    H2
        |     |
       -B1    B4-


    .. plot::
      :include-source:

      import pp

      c = pp.c.mzi_arm()
      pp.plotgds(c)


    """
    if not with_elec_connections:
        straight_heater_factory = straight_factory

    _bend = bend90_factory(radius=bend_radius)

    straight_vheater = straight_heater_factory(length=L0)
    straight_h = straight_factory(length=L_top)
    straight_v = straight_factory(length=DL) if DL > 0 else None

    port_number = 1 if with_elec_connections else 0

    string_to_device_in_out_ports = {
        "A": (_bend, "W0", "N0"),
        "B": (_bend, "N0", "W0"),
        "H": (straight_vheater, f"W{port_number}", f"E{port_number}"),
        "Sh": (straight_h, "W0", "E0"),
        "Sv": (straight_v, "W0", "E0"),
    }

    sequence = ["A", "Sv", "H", "B", "Sh", "B", "H", "Sv", "A"]

    if with_elec_connections:
        ports_map = {
            "E_0": ("H2", "E_1"),
            "E_1": ("H1", "E_0"),
            "E_2": ("H1", "E_1"),
            "E_3": ("H2", "E_0"),
        }
    else:
        ports_map = {}

    component = component_sequence(
        sequence,
        string_to_device_in_out_ports,
        ports_map,
        input_port_name="W0",
        output_port_name="E0",
    )
    return component


@autoname
def mzi2x2(
    CL_1: float = 20.147,
    L0: float = 60.0,
    DL: float = 7.38,
    L2: float = 10.0,
    gap: float = 0.234,
    bend_radius: float = 10.0,
    bend90_factory: Callable = bend_circular,
    straight_heater_factory: Callable = waveguide_heater,
    straight_factory: Callable = waveguide,
    coupler_factory: Callable = coupler,
    with_elec_connections: bool = False,
) -> Component:
    """ Mzi 2x2

    Args:
        CL_1: coupler length
        L0: vertical length for both and top arms
        DL: bottom arm extra length
        L2: L_top horizontal length
        gap: 0.235
        bend_radius: 10.0
        bend90_factory: bend_circular
        straight_heater_factory: waveguide_heater or waveguide
        straight_factory: waveguide
        coupler_factory: coupler


    .. code::

         __L2__
        |      |
        L0     L0
        |      |
      ==|      |==
        |      |
        L0     L0
        |      |
        DL     DL
        |      |
        |__L2__|


    .. code::

               top_arm
        ==CL_1=       =CL_1===
               bot_arm


    .. plot::
      :include-source:

      import pp

      c = pp.c.mzi2x2(CL_1=10., gap=0.2)
      pp.plotgds(c)

    """
    if not with_elec_connections:
        straight_heater_factory = straight_factory

    if callable(coupler_factory):
        cpl = coupler_factory(length=CL_1, gap=gap)
    else:
        cpl = coupler_factory

    arm_defaults = {
        "L_top": L2,
        "bend_radius": bend_radius,
        "bend90_factory": bend90_factory,
        "straight_heater_factory": straight_heater_factory,
        "straight_factory": straight_factory,
        "with_elec_connections": with_elec_connections,
    }

    arm_top = mzi_arm(L0=L0, **arm_defaults)
    arm_bot = mzi_arm(L0=L0, DL=DL, **arm_defaults)

    components = {
        "CP1": (cpl, "None"),
        "CP2": (cpl, "None"),
        "arm_top": (arm_top, "None"),
        "arm_bot": (arm_bot, "mirror_x"),
    }

    connections = [
        ## Top arm
        ("CP1", "E1", "arm_top", "W0"),
        ("arm_top", "E0", "CP2", "W1"),
        ## Bottom arm
        ("CP1", "E0", "arm_bot", "W0"),
        ("arm_bot", "E0", "CP2", "W0"),
    ]

    if with_elec_connections:

        ports_map = {
            "W0": ("CP1", "W0"),
            "W1": ("CP1", "W1"),
            "E0": ("CP2", "E0"),
            "E1": ("CP2", "E1"),
            "E_TOP_0": ("arm_top", "E_0"),
            "E_TOP_1": ("arm_top", "E_1"),
            "E_TOP_2": ("arm_top", "E_2"),
            "E_TOP_3": ("arm_top", "E_3"),
            "E_BOT_0": ("arm_bot", "E_0"),
            "E_BOT_1": ("arm_bot", "E_1"),
            "E_BOT_2": ("arm_bot", "E_2"),
            "E_BOT_3": ("arm_bot", "E_3"),
        }

        component = netlist_to_component(components, connections, ports_map)
        # Need to connect common ground and redefine electrical ports

        ports = component.ports
        y_elec = ports["E_TOP_0"].y
        for ls, le in [
            ("E_BOT_0", "E_BOT_1"),
            ("E_TOP_0", "E_TOP_1"),
            ("E_BOT_2", "E_TOP_2"),
        ]:
            component.add_polygon(line(ports[ls], ports[le]), layer=ports[ls].layer)

        # Add GND
        component.add_port(
            name="GND",
            midpoint=0.5 * (ports["E_BOT_2"].midpoint + ports["E_TOP_2"].midpoint),
            orientation=180,
            width=ports["E_BOT_2"].width,
            layer=ports["E_BOT_2"].layer,
            port_type="dc",
        )

        component.ports["E_TOP_3"].orientation = 0
        component.ports["E_BOT_3"].orientation = 0

        # Remove the eletrical ports that we have just used internally
        for lbl in ["E_BOT_0", "E_BOT_1", "E_TOP_0", "E_TOP_1", "E_BOT_2", "E_TOP_2"]:
            component.ports.pop(lbl)

        # Reroute electrical ports
        _e_ports = select_electrical_ports(component)
        conn, e_ports = route_elec_ports_to_side(_e_ports, side="north", y=y_elec)

        for c in conn:
            component.add(c)

        for p in e_ports:
            component.ports[p.name] = p

        # Create nice electrical port names
        component.ports["HT1"] = component.ports["E_TOP_3"]
        component.ports.pop("E_TOP_3")

        component.ports["HT2"] = component.ports["E_BOT_3"]
        component.ports.pop("E_BOT_3")

        # Make sure each port knows its name
        for k, p in component.ports.items():
            p.name = k

    else:
        ports_map = {
            "W0": ("CP1", "W0"),
            "W1": ("CP1", "W1"),
            "E0": ("CP2", "E0"),
            "E1": ("CP2", "E1"),
        }

        component = netlist_to_component(components, connections, ports_map)

    return component


def get_mzi_delta_length(m, neff=2.4, wavelength=1.55):
    """ m*wavelength = neff * delta_length """
    return m * wavelength / neff


if __name__ == "__main__":
    import pp

    # print(get_mzi_delta_length(m=15))
    # print(get_mzi_delta_length(m=150))

    c = mzi2x2(with_elec_connections=True, pins=True)
    # for p in c.ports.values():
    #     print(p.port_type)

    # c = mzi_arm(DL=100)
    # c = mzi2x2(straight_heater_factory=waveguide_heater, with_elec_connections=True)
    # pp.write_gds(c, "mzi.gds")
    # print(c)
    # print(hash(frozenset(c.settings.items())))
    # print(hash(c))

    pp.write_gds(c, pp.CONFIG["gdsdir"] / "mzi2x2.gds")
    pp.show(c)



from setuptools import find_packages, setup


def get_install_requires():
    with open("requirements.txt", "r") as f:
        return [line.strip() for line in f.readlines() if not line.startswith("-")]


setup(
    name="gdsfactory",
    url="https://github.com/gdsfactory/gdsfactory",
    version="1.4.2",
    author="PsiQ",
    scripts=["pp/pf.py"],
    description="python libraries to generate GDS layouts",
    packages=find_packages(),
    include_package_data=True,
    install_requires=get_install_requires(),
    python_requires=">=3.6",
    entry_points="""
        [console_scripts]
        pf=pp.pf:cli
    """,
)



""" pp Photonics package

functions:

    - pp.show(): writes and shows the GDS in Klayout using klive
    - pp.plotgds(): plots GDS in matplotlib
    - pp.import_gds(): returns a Component from a GDS

classes:

    - pp.Component
    - pp.Port
    - CONFIG

modules:

    - c: components
    - routing
    - layer: GDS layers
"""
from phidl import quickplot as qp
import phidl.geometry as pg

# NOTE: import order matters. Only change the order if you know what you are doing
from pp.config import CONFIG, call_if_func, conf
from pp.component import Component, ComponentReference
from pp.port import Port
from pp.name import autoname
from pp.layers import LAYER
from pp.load_component import load_component
from pp.load_csv import load_csv

from pp.write_component import get_component_type
from pp.write_component import show
from pp.write_component import write_gds
from pp.write_component import write_component_type
from pp.write_component import write_component
from pp.write_doe import write_doe

import pp.components as c
import pp.routing as routing
import pp.bias as bias
import pp.klive as klive
import pp.sp as sp
import pp.port as port
import pp.units as units

from pp.component_from_yaml import component_from_yaml

from pp.components import component_factory
from pp.components.extension import extend_port
from pp.components.extension import extend_ports
from pp.add_padding import add_padding
from pp.add_pins import add_pins
from pp.import_gds import import_gds
from pp.import_phidl_component import import_phidl_component
from pp.plotgds import plotgds
from pp.pack import pack
from pp.boolean import boolean


__all__ = [
    "CONFIG",
    "LAYER",
    "Component",
    "ComponentReference",
    "bias",
    "autoname",
    "add_padding",
    "add_pins",
    "import_gds",
    "import_phidl_component",
    "c",
    "conf",
    "component_factory",
    "call_if_func",
    "extend_port",
    "extend_ports",
    "boolean",
    "get_component_type",
    "klive",
    "load_component",
    "load_csv",
    "plotgds",
    "pack",
    "qp",
    "pg",
    "port",
    "routing",
    "show",
    "sp",
    "write_component",
    "write_component_type",
    "write_doe",
    "write_gds",
    "Port",
    "component_from_yaml",
    "units",
]
__version__ = "1.4.2"


if __name__ == "__main__":
    print(__all__)



""" From phidl tutorial

# Flattening a Component

Sometimes you want to remove cell structure from a Component while keeping all
of the shapes/polygons intact and in place.  The c.flatten() keeps all the
polygons in c, but removes all the underlying references it's attached to.
Also, if you specify the `single_layer` argument it will move all of the
polyons to that single layer.

"""

import pp

c = pp.Component("waveguides_sample")


wg1 = c << pp.c.waveguide(length=10, width=1)
wg2 = c << pp.c.waveguide(length=10, width=2, layer=pp.LAYER.SLAB90)
wg3 = c << pp.c.waveguide(length=10, width=3, layer=pp.LAYER.SLAB150)

wg2.connect(port="W0", destination=wg1.ports["E0"])
wg3.connect(port="W0", destination=wg2.ports["E0"], overlap=1)


c.flatten()
# c.flatten(single_layer = pp.LAYER.WG) # you can also flatten all polygons into a single layer

pp.qp(c)  # quickplot it!
pp.show(c)  # show it in klayout



""" design rule checking """
import numpy as np
from numpy import bool_, float64
from typing import Union


def on_grid(x: float64, nm: int = 1) -> bool_:
    return np.isclose(snap_to_grid(x, nm=nm), x)


def on_1nm_grid(x):
    return np.isclose(snap_to_1nm_grid(x), x)


def on_2nm_grid(x):
    return np.isclose(snap_to_2nm_grid(x), x)


def assert_on_1nm_grid(x: float) -> None:
    assert np.isclose(snap_to_1nm_grid(x), x), f"{x} needs to be on 1nm grid"


def assert_on_2nm_grid(x: float) -> None:
    assert np.isclose(snap_to_2nm_grid(x), x), f"{x} needs to be on 1nm grid"


def snap_to_grid(x: Union[float64, float], nm: int = 1) -> float64:
    return nm * np.round(np.array(x) * 1e3 / nm) / 1e3


def snap_to_1nm_grid(x: float) -> float64:
    return snap_to_grid(x, nm=1)


def snap_to_2nm_grid(x: float) -> float64:
    return snap_to_grid(x, nm=2)


def snap_to_5nm_grid(x: float64) -> float64:
    return snap_to_grid(x, nm=5)


def test_snap_to_1nm_grid():
    assert snap_to_1nm_grid(1.1e-3) == 0.001


def test_snap_to_2nm_grid():
    assert snap_to_2nm_grid(1.1e-3) == 0.002
    assert snap_to_2nm_grid(3.1e-3) == 0.004


def test_on_1nm_grid():
    assert not on_1nm_grid(1.1e-3)
    assert on_1nm_grid(1e-3)


def test_on_2nm_grid():
    assert not on_2nm_grid(1.1e-3)
    assert not on_2nm_grid(1e-3)
    assert on_2nm_grid(2e-3)


if __name__ == "__main__":
    # print(snap_to_1nm_grid(1.1e-3))
    # print(snap_to_2nm_grid(1.1e-3))
    # print(snap_to_2nm_grid(3.1e-3))

    # print(on_1nm_grid(1.1e-3))
    # print(on_1nm_grid(1e-3))

    print(on_2nm_grid(1.1e-3))
    print(on_2nm_grid(1e-3))
    print(on_2nm_grid(2e-3))



from typing import Tuple
import pp
from pp import components as pc
from pp.component import Component


@pp.autoname
def _via_iterable(
    via_spacing: int,
    wire_width: int,
    wiring1_layer: int,
    wiring2_layer: int,
    via_layer: int,
    via_width: int,
) -> Component:
    VI = pp.Component()
    wire1 = VI.add_ref(pc.compass(size=(via_spacing, wire_width), layer=wiring1_layer))
    wire2 = VI.add_ref(pc.compass(size=(via_spacing, wire_width), layer=wiring2_layer))
    via1 = VI.add_ref(pc.compass(size=(via_width, via_width), layer=via_layer))
    via2 = VI.add_ref(pc.compass(size=(via_width, via_width), layer=via_layer))
    wire1.connect(port="E", destination=wire2.ports["W"], overlap=wire_width)
    via1.connect(
        port="W", destination=wire1.ports["E"], overlap=(wire_width + via_width) / 2
    )
    via2.connect(
        port="W", destination=wire2.ports["E"], overlap=(wire_width + via_width) / 2
    )
    VI.add_port(name="W", port=wire1.ports["W"])
    VI.add_port(name="E", port=wire2.ports["E"])
    VI.add_port(
        name="S",
        midpoint=[(1 * wire_width) + wire_width / 2, -wire_width / 2],
        width=wire_width,
        orientation=-90,
    )
    VI.add_port(
        name="N",
        midpoint=[(1 * wire_width) + wire_width / 2, wire_width / 2],
        width=wire_width,
        orientation=90,
    )

    return VI


@pp.autoname
def test_via(
    num_vias: int = 100,
    wire_width: int = 10,
    via_width: int = 15,
    via_spacing: int = 40,
    pad_size: Tuple[int, int] = (300, 300),
    min_pad_spacing: int = 0,
    pad_layer: int = 0,
    wiring1_layer: int = 1,
    wiring2_layer: int = 2,
    via_layer: int = 3,
) -> Component:
    """ Via cutback to extract via resistance
    from phidl.geometry

    Args:
        num_vias=100
        wire_width=10
        via_width=15
        via_spacing=40
        pad_size=(300, 300)
        min_pad_spacing=0
        pad_layer=0
        wiring1_layer=1
        wiring2_layer=2
        via_layer=3

    Usage:
        Call via_route_test_structure() by indicating the number of vias you want drawn. You can also change the other parameters however
        if you do not specifiy a value for a parameter it will just use the default value
        Ex::

            via_route_test_structure(num_vias=54)

        - or -::

            via_route_test_structure(num_vias=12, pad_size=(100,100),wire_width=8)

        total requested vias (num_vias) -> this needs to be even
        pad size (pad_size) -> given in a pair (width, height)
        wire_width -> how wide each wire should be
        pad_layer -> GDS layer number of the pads
        wiring1_layer -> GDS layer number of the top wiring
        wiring2_layer -> GDS layer number of the bottom wiring
        via_layer -> GDS layer number of the vias
        ex: via_route(54, min_pad_spacing=300)

    .. plot::
      :include-source:

      import pp

      c = pp.c.test_via()
      pp.plotgds(c)
    """

    VR = pp.Component()
    pad1 = VR.add_ref(pc.rectangle(size=pad_size, layer=pad_layer))
    pad1_overlay = VR.add_ref(pc.rectangle(size=pad_size, layer=wiring1_layer))
    pad2 = VR.add_ref(pc.rectangle(size=pad_size, layer=pad_layer))
    pad2_overlay = VR.add_ref(pc.rectangle(size=pad_size, layer=wiring1_layer))
    nub = VR.add_ref(pc.compass(size=(3 * wire_width, wire_width), layer=pad_layer))
    nub_overlay = VR.add_ref(
        pc.compass(size=(3 * wire_width, wire_width), layer=wiring1_layer)
    )
    head = VR.add_ref(pc.compass(size=(wire_width, wire_width), layer=pad_layer))
    head_overlay = VR.add_ref(
        pc.compass(size=(wire_width, wire_width), layer=wiring1_layer)
    )
    nub.ymax = pad1.ymax - 5
    nub.xmin = pad1.xmax
    nub_overlay.ymax = pad1.ymax - 5
    nub_overlay.xmin = pad1.xmax
    head.connect(port="W", destination=nub.ports["E"])
    head_overlay.connect(port="W", destination=nub_overlay.ports["E"])
    pad1_overlay.xmin = pad1.xmin
    pad1_overlay.ymin = pad1.ymin

    old_port = head.ports["S"]
    count = 0
    width_via_iter = 2 * via_spacing - 2 * wire_width

    pad2.xmin = pad1.xmax + min_pad_spacing
    up = False
    down = True
    edge = True
    current_width = 3 * wire_width + wire_width  # width of nub and 1 overlap
    obj_old = head
    obj = head
    via_iterable = _via_iterable(
        via_spacing=via_spacing,
        wire_width=wire_width,
        wiring1_layer=wiring1_layer,
        wiring2_layer=wiring2_layer,
        via_layer=via_layer,
        via_width=via_width,
    )
    while (count + 2) <= num_vias:
        obj = VR.add_ref(via_iterable)
        obj.connect(port="W", destination=old_port, overlap=wire_width)
        old_port = obj.ports["E"]
        edge = False
        if obj.ymax > pad1.ymax:
            obj.connect(port="W", destination=obj_old.ports["S"], overlap=wire_width)
            old_port = obj.ports["S"]
            current_width += width_via_iter
            down = True
            up = False
            edge = True

        elif obj.ymin < pad1.ymin:
            obj.connect(port="W", destination=obj_old.ports["N"], overlap=wire_width)
            old_port = obj.ports["N"]
            current_width += width_via_iter
            up = True
            down = False
            edge = True
        count = count + 2
        obj_old = obj

    if (
        current_width < min_pad_spacing
        and (min_pad_spacing - current_width) > 3 * wire_width
    ):
        tail = VR.add_ref(
            pc.compass(
                size=(min_pad_spacing - current_width + wire_width, wire_width),
                layer=wiring1_layer,
            )
        )
        tail_overlay = VR.add_ref(
            pc.compass(
                size=(min_pad_spacing - current_width + wire_width, wire_width),
                layer=pad_layer,
            )
        )
    else:
        tail = VR.add_ref(
            pc.compass(size=(3 * wire_width, wire_width), layer=wiring1_layer)
        )
        tail_overlay = VR.add_ref(
            pc.compass(size=(3 * wire_width, wire_width), layer=wiring1_layer)
        )

    if up and not edge:
        tail.connect(port="W", destination=obj.ports["S"], overlap=wire_width)
        tail_overlay.connect(port="W", destination=obj.ports["S"], overlap=wire_width)
    elif down and not edge:
        tail.connect(port="W", destination=obj.ports["N"], overlap=wire_width)
        tail_overlay.connect(port="W", destination=obj.ports["N"], overlap=wire_width)
    else:
        tail.connect(port="W", destination=obj.ports["E"], overlap=wire_width)
        tail_overlay.connect(port="W", destination=obj.ports["E"], overlap=wire_width)

    pad2.xmin = tail.xmax
    pad2_overlay.xmin = pad2.xmin
    pad2_overlay.ymin = pad2.ymin

    return VR


if __name__ == "__main__":
    c = test_via()
    pp.show(c)
